# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.save` function."""

import cftime
import cf_units

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.coords import DimCoord, CellMethod
from iris.fileformats._ff_cross_references import STASH_TRANS
import iris.fileformats.pp as pp
from iris.fileformats.pp_save_rules import _lbproc_rules, verify
import iris.tests.stock as stock


def _pp_save_ppfield_values(cube):
    """
    Emulate saving a cube as PP, and capture the resulting PP field values.

    """
    # Create a test object to stand in for a real PPField.
    pp_field = mock.MagicMock(spec=pp.PPField3)
    # Add minimal content required by the pp.save operation.
    pp_field.HEADER_DEFN = pp.PPField3.HEADER_DEFN
    # Save cube to a dummy file, mocking the internally created PPField
    with mock.patch("iris.fileformats.pp.PPField3", return_value=pp_field):
        target_filelike = mock.Mock(name="target")
        target_filelike.mode = "b"
        pp.save(cube, target_filelike)
    # Return pp-field mock with all the written properties
    return pp_field


class TestVertical(tests.IrisTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()

    def test_pseudo_level(self):
        pseudo_level = 123
        coord = DimCoord(pseudo_level, long_name="pseudo_level", units="1")
        self.cube.add_aux_coord(coord)
        lbuser5_produced = _pp_save_ppfield_values(self.cube).lbuser[4]
        self.assertEqual(pseudo_level, lbuser5_produced)

    def test_soil_level(self):
        soil_level = 314
        coord = DimCoord(soil_level, long_name="soil_model_level_number")
        self.cube.add_aux_coord(coord)
        self.cube.standard_name = "moisture_content_of_soil_layer"
        field = _pp_save_ppfield_values(self.cube)
        self.assertEqual(field.lbvc, 6)
        self.assertEqual(field.lblev, soil_level)
        self.assertEqual(field.blev, soil_level)
        self.assertEqual(field.brsvd[0], 0)
        self.assertEqual(field.brlev, 0)

    def test_soil_depth(self):
        lower, point, upper = 1, 2, 3
        coord = DimCoord(point, standard_name="depth", bounds=[[lower, upper]])
        self.cube.add_aux_coord(coord)
        self.cube.standard_name = "moisture_content_of_soil_layer"
        field = _pp_save_ppfield_values(self.cube)
        self.assertEqual(field.lbvc, 6)
        self.assertEqual(field.lblev, 0)
        self.assertEqual(field.blev, point)
        self.assertEqual(field.brsvd[0], lower)
        self.assertEqual(field.brlev, upper)


class TestLbfcProduction(tests.IrisTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()

    def check_cube_stash_yields_lbfc(self, stash, lbfc_expected):
        if stash:
            self.cube.attributes["STASH"] = stash
        lbfc_produced = _pp_save_ppfield_values(self.cube).lbfc
        self.assertEqual(lbfc_produced, lbfc_expected)

    def test_known_stash(self):
        stashcode_str = "m04s07i002"
        self.assertIn(stashcode_str, STASH_TRANS)
        self.check_cube_stash_yields_lbfc(stashcode_str, 359)

    def test_unknown_stash(self):
        stashcode_str = "m99s99i999"
        self.assertNotIn(stashcode_str, STASH_TRANS)
        self.check_cube_stash_yields_lbfc(stashcode_str, 0)

    def test_no_stash(self):
        self.assertNotIn("STASH", self.cube.attributes)
        self.check_cube_stash_yields_lbfc(None, 0)

    def check_cube_name_units_yields_lbfc(self, name, units, lbfc_expected):
        self.cube.rename(name)
        self.cube.units = units
        lbfc_produced = _pp_save_ppfield_values(self.cube).lbfc
        self.assertEqual(
            lbfc_produced,
            lbfc_expected,
            "Lbfc for ({!r} / {!r}) should be {:d}, "
            "got {:d}".format(name, units, lbfc_expected, lbfc_produced),
        )

    def test_name_units_to_lbfc(self):
        # Check LBFC value produced from name and units.
        self.check_cube_name_units_yields_lbfc("sea_ice_temperature", "K", 209)

    def test_bad_name_units_to_lbfc_0(self):
        # Check that badly-formed / unrecognised cases yield LBFC == 0.
        self.check_cube_name_units_yields_lbfc(
            "sea_ice_temperature", "degC", 0
        )
        self.check_cube_name_units_yields_lbfc("Junk_Name", "K", 0)


class TestLbsrceProduction(tests.IrisTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()

    def check_cube_um_source_yields_lbsrce(
        self, source_str=None, um_version_str=None, lbsrce_expected=None
    ):
        if source_str is not None:
            self.cube.attributes["source"] = source_str
        if um_version_str is not None:
            self.cube.attributes["um_version"] = um_version_str
        lbsrce_produced = _pp_save_ppfield_values(self.cube).lbsrce
        self.assertEqual(lbsrce_produced, lbsrce_expected)

    def test_none(self):
        self.check_cube_um_source_yields_lbsrce(None, None, 0)

    def test_source_only_no_version(self):
        self.check_cube_um_source_yields_lbsrce(
            "Data from Met Office Unified Model", None, 1111
        )

    def test_source_only_with_version(self):
        self.check_cube_um_source_yields_lbsrce(
            "Data from Met Office Unified Model 12.17", None, 12171111
        )

    def test_um_version(self):
        self.check_cube_um_source_yields_lbsrce(
            "Data from Met Office Unified Model 12.17", "25.36", 25361111
        )


class Test_Save__LbprocProduction(tests.IrisTest):
    # This test class is a little different to the others.
    # If it called `pp.save` via `_pp_save_ppfield_values` it would run
    # `pp_save_rules.verify` and run all the save rules. As this class uses
    # a 3D cube with a time coord it would run the time rules, which would fail
    # because the mock object does not set up the `pp.lbtim` attribute
    # correctly (i.e. as a `SplittableInt` object).
    # To work around this we call the lbproc rules directly here.

    def setUp(self):
        self.cube = stock.realistic_3d()
        self.pp_field = mock.MagicMock(spec=pp.PPField3)
        self.pp_field.HEADER_DEFN = pp.PPField3.HEADER_DEFN
        self.patch("iris.fileformats.pp.PPField3", return_value=self.pp_field)

    def test_no_cell_methods(self):
        lbproc = _lbproc_rules(self.cube, self.pp_field).lbproc
        self.assertEqual(lbproc, 0)

    def test_mean(self):
        self.cube.cell_methods = (CellMethod("mean", "time", "1 hour"),)
        lbproc = _lbproc_rules(self.cube, self.pp_field).lbproc
        self.assertEqual(lbproc, 128)

    def test_minimum(self):
        self.cube.cell_methods = (CellMethod("minimum", "time", "1 hour"),)
        lbproc = _lbproc_rules(self.cube, self.pp_field).lbproc
        self.assertEqual(lbproc, 4096)

    def test_maximum(self):
        self.cube.cell_methods = (CellMethod("maximum", "time", "1 hour"),)
        lbproc = _lbproc_rules(self.cube, self.pp_field).lbproc
        self.assertEqual(lbproc, 8192)


class TestTimeMean(tests.IrisTest):
    """
    Tests that time mean cell method is converted to pp appropriately.

    Pattern is pairs of tests - one with time mean method, and one without, to
    show divergent behaviour.

    """

    def test_t1_time_mean(self):
        cube = _get_single_time_cube(set_time_mean=True)
        tc = cube.coord(axis="t")
        expected = tc.units.num2date(0)

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.t1

        self.assertEqual(expected, actual)

    def test_t1_no_time_mean(self):
        cube = _get_single_time_cube()
        tc = cube.coord(axis="t")
        expected = tc.units.num2date(15)

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.t1

        self.assertEqual(expected, actual)

    def test_t2_time_mean(self):
        cube = _get_single_time_cube(set_time_mean=True)
        tc = cube.coord(axis="t")
        expected = tc.units.num2date(30)

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.t2

        self.assertEqual(expected, actual)

    def test_t2_no_time_mean(self):
        cube = _get_single_time_cube(set_time_mean=False)
        expected = cftime.datetime(0, 0, 0)

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.t2
        self.assertEqual(expected, actual)

    def test_lbft_no_forecast_time(self):
        # Different pattern here: checking that lbft hasn't been changed from
        # the default value.
        cube = _get_single_time_cube()
        mock_lbft = mock.sentinel.lbft

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            pp_field.lbft = mock_lbft
            verify(cube, pp_field)
        actual = pp_field.lbft

        assert mock_lbft is actual

    def test_lbtim_no_time_mean(self):
        cube = _get_single_time_cube()
        expected_ib = 0
        expected_ic = 2  # 360 day calendar

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual_ib = pp_field.lbtim.ib
        actual_ic = pp_field.lbtim.ic

        self.assertEqual(expected_ib, actual_ib)
        self.assertEqual(expected_ic, actual_ic)

    def test_lbtim_time_mean(self):
        cube = _get_single_time_cube(set_time_mean=True)
        expected_ib = 2  # Time mean
        expected_ic = 2  # 360 day calendar

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual_ib = pp_field.lbtim.ib
        actual_ic = pp_field.lbtim.ic

        self.assertEqual(expected_ib, actual_ib)
        self.assertEqual(expected_ic, actual_ic)

    def test_lbproc_no_time_mean(self):
        cube = _get_single_time_cube()
        expected = 0

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.lbproc

        self.assertEqual(expected, actual)

    def test_lbproc_time_mean(self):
        cube = _get_single_time_cube(set_time_mean=True)
        expected = 128

        with mock.patch(
            "iris.fileformats.pp.PPField3", autospec=True
        ) as pp_field:
            verify(cube, pp_field)
        actual = pp_field.lbproc

        self.assertEqual(expected, actual)


def _get_single_time_cube(set_time_mean=False):
    cube = stock.realistic_3d()[0:1, :, :]
    cube.remove_coord("time")
    cube.remove_coord("forecast_period")
    tc = DimCoord(
        points=[15],
        standard_name="time",
        units=cf_units.Unit("days since epoch", calendar="360_day"),
        bounds=[[0, 30]],
    )
    cube.add_dim_coord(tc, 0)
    if set_time_mean:
        cube.cell_methods = (CellMethod("mean", coords="time"),)
    return cube


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.RMS` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_lazy_data
from iris.analysis import RMS


class Test_aggregate(tests.IrisTest):
    def test_1d(self):
        # 1-dimensional input
        data = np.array([5, 2, 6, 4], dtype=np.float64)
        rms = RMS.aggregate(data, 0)
        expected_rms = 4.5
        self.assertAlmostEqual(rms, expected_rms)

    def test_2d(self):
        # 2-dimensional input
        data = np.array([[5, 2, 6, 4], [12, 4, 10, 8]], dtype=np.float64)
        expected_rms = np.array([4.5, 9.0], dtype=np.float64)
        rms = RMS.aggregate(data, 1)
        self.assertArrayAlmostEqual(rms, expected_rms)

    def test_1d_weighted(self):
        # 1-dimensional input with weights
        data = np.array([4, 7, 10, 8], dtype=np.float64)
        weights = np.array([1, 4, 3, 2], dtype=np.float64)
        expected_rms = 8.0
        rms = RMS.aggregate(data, 0, weights=weights)
        self.assertAlmostEqual(rms, expected_rms)

    def test_2d_weighted(self):
        # 2-dimensional input with weights
        data = np.array([[4, 7, 10, 8], [14, 16, 20, 8]], dtype=np.float64)
        weights = np.array([[1, 4, 3, 2], [2, 1, 1.5, 0.5]], dtype=np.float64)
        expected_rms = np.array([8.0, 16.0], dtype=np.float64)
        rms = RMS.aggregate(data, 1, weights=weights)
        self.assertArrayAlmostEqual(rms, expected_rms)

    def test_unit_weighted(self):
        # unit weights should be the same as no weights
        data = np.array([5, 2, 6, 4], dtype=np.float64)
        weights = np.ones_like(data)
        rms = RMS.aggregate(data, 0, weights=weights)
        expected_rms = 4.5
        self.assertAlmostEqual(rms, expected_rms)

    def test_masked(self):
        # masked entries should be completely ignored
        data = ma.array(
            [5, 10, 2, 11, 6, 4],
            mask=[False, True, False, True, False, False],
            dtype=np.float64,
        )
        expected_rms = 4.5
        rms = RMS.aggregate(data, 0)
        self.assertAlmostEqual(rms, expected_rms)

    def test_masked_weighted(self):
        # weights should work properly with masked arrays
        data = ma.array(
            [4, 7, 18, 10, 11, 8],
            mask=[False, False, True, False, True, False],
            dtype=np.float64,
        )
        weights = np.array([1, 4, 5, 3, 8, 2], dtype=np.float64)
        expected_rms = 8.0
        rms = RMS.aggregate(data, 0, weights=weights)
        self.assertAlmostEqual(rms, expected_rms)


class Test_lazy_aggregate(tests.IrisTest):
    def test_1d(self):
        # 1-dimensional input.
        data = as_lazy_data(np.array([5, 2, 6, 4], dtype=np.float64))
        rms = RMS.lazy_aggregate(data, 0)
        expected_rms = 4.5
        self.assertAlmostEqual(rms, expected_rms)

    def test_2d(self):
        # 2-dimensional input.
        data = as_lazy_data(
            np.array([[5, 2, 6, 4], [12, 4, 10, 8]], dtype=np.float64)
        )
        expected_rms = np.array([4.5, 9.0], dtype=np.float64)
        rms = RMS.lazy_aggregate(data, 1)
        self.assertArrayAlmostEqual(rms, expected_rms)

    def test_1d_weighted(self):
        # 1-dimensional input with weights.
        data = as_lazy_data(np.array([4, 7, 10, 8], dtype=np.float64))
        weights = np.array([1, 4, 3, 2], dtype=np.float64)
        expected_rms = 8.0
        # https://github.com/dask/dask/issues/3846.
        with self.assertRaisesRegex(TypeError, "unexpected keyword argument"):
            rms = RMS.lazy_aggregate(data, 0, weights=weights)
            self.assertAlmostEqual(rms, expected_rms)

    def test_1d_lazy_weighted(self):
        # 1-dimensional input with lazy weights.
        data = as_lazy_data(np.array([4, 7, 10, 8], dtype=np.float64))
        weights = as_lazy_data(np.array([1, 4, 3, 2], dtype=np.float64))
        expected_rms = 8.0
        # https://github.com/dask/dask/issues/3846.
        with self.assertRaisesRegex(TypeError, "unexpected keyword argument"):
            rms = RMS.lazy_aggregate(data, 0, weights=weights)
            self.assertAlmostEqual(rms, expected_rms)

    def test_2d_weighted(self):
        # 2-dimensional input with weights.
        data = as_lazy_data(
            np.array([[4, 7, 10, 8], [14, 16, 20, 8]], dtype=np.float64)
        )
        weights = np.array([[1, 4, 3, 2], [2, 1, 1.5, 0.5]], dtype=np.float64)
        expected_rms = np.array([8.0, 16.0], dtype=np.float64)
        # https://github.com/dask/dask/issues/3846.
        with self.assertRaisesRegex(TypeError, "unexpected keyword argument"):
            rms = RMS.lazy_aggregate(data, 1, weights=weights)
            self.assertArrayAlmostEqual(rms, expected_rms)

    def test_unit_weighted(self):
        # Unit weights should be the same as no weights.
        data = as_lazy_data(np.array([5, 2, 6, 4], dtype=np.float64))
        weights = np.ones_like(data)
        expected_rms = 4.5
        # https://github.com/dask/dask/issues/3846.
        with self.assertRaisesRegex(TypeError, "unexpected keyword argument"):
            rms = RMS.lazy_aggregate(data, 0, weights=weights)
            self.assertAlmostEqual(rms, expected_rms)

    def test_masked(self):
        # Masked entries should be completely ignored.
        data = as_lazy_data(
            ma.array(
                [5, 10, 2, 11, 6, 4],
                mask=[False, True, False, True, False, False],
                dtype=np.float64,
            )
        )
        expected_rms = 4.5
        rms = RMS.lazy_aggregate(data, 0)
        self.assertAlmostEqual(rms, expected_rms)

    def test_masked_weighted(self):
        # Weights should work properly with masked arrays, but currently don't
        # (see https://github.com/dask/dask/issues/3846).
        # For now, masked weights are simply not supported.
        data = as_lazy_data(
            ma.array(
                [4, 7, 18, 10, 11, 8],
                mask=[False, False, True, False, True, False],
                dtype=np.float64,
            )
        )
        weights = np.array([1, 4, 5, 3, 8, 2])
        expected_rms = 8.0
        with self.assertRaisesRegex(TypeError, "unexpected keyword argument"):
            rms = RMS.lazy_aggregate(data, 0, weights=weights)
            self.assertAlmostEqual(rms, expected_rms)


class Test_name(tests.IrisTest):
    def test(self):
        self.assertEqual(RMS.name(), "root_mean_square")


class Test_aggregate_shape(tests.IrisTest):
    def test(self):
        shape = ()
        kwargs = dict()
        self.assertTupleEqual(RMS.aggregate_shape(**kwargs), shape)
        kwargs = dict(tom="jerry", calvin="hobbes")
        self.assertTupleEqual(RMS.aggregate_shape(**kwargs), shape)


if __name__ == "__main__":
    tests.main()



def sample_routine(arg1, arg2, kwarg1="foo", kwarg2=None):
    """
    Purpose section text goes here.

    Description section longer text goes here.

    Args:

    * arg1 (numpy.ndarray):
        First argument description.
    * arg2 (numpy.ndarray):
        Second argument description.

    Kwargs:

    * kwarg1 (string):
        The first keyword argument. This argument description
        can be multi-lined.
    * kwarg2 (Boolean or None):
        The second keyword argument.

    Returns:
        numpy.ndarray of arg1 * arg2

    """
    pass



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coords.AncillaryVariable` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import dask.array as da
import numpy as np
import numpy.ma as ma

from iris.tests.unit.coords import CoordTestMixin, lazyness_string

from cf_units import Unit
from iris.coords import AncillaryVariable
from iris.cube import Cube
from iris._lazy_data import as_lazy_data


def data_all_dtypes_and_lazynesses(self):
    # Generate ancillary variables with real and lazy data, and a few different
    # dtypes.
    data_types = ["real", "lazy"]
    dtypes = [np.int16, np.int32, np.float32, np.float64]
    for dtype in dtypes:
        for data_type_name in data_types:
            data = np.asarray(self.data_real, dtype=dtype)
            if data_type_name == "lazy":
                data = da.from_array(data, data.shape)
            ancill_var = AncillaryVariable(data)
            result = (ancill_var, data_type_name)
            yield result


class AncillaryVariableTestMixin(CoordTestMixin):
    # Define a 2-D default array shape.
    def setupTestArrays(self, shape=(2, 3), masked=False):
        # Create concrete and lazy data test arrays, given a desired shape.
        # If masked=True, also add masked arrays with some or no masked data.
        n_vals = np.prod(shape)
        # Note: the values must be integral for testing integer dtypes.
        values = 100.0 + 10.0 * np.arange(n_vals, dtype=float).reshape(shape)
        self.data_real = values
        self.data_lazy = da.from_array(values, values.shape)

        if masked:
            mvalues = ma.array(values)
            self.no_masked_data_real = mvalues
            self.no_masked_data_lazy = da.from_array(
                mvalues, mvalues.shape, asarray=False
            )
            mvalues = ma.array(mvalues, copy=True)
            mvalues[0] = ma.masked
            self.masked_data_real = mvalues
            self.masked_data_lazy = da.from_array(
                mvalues, mvalues.shape, asarray=False
            )


class Test__init__(tests.IrisTest, AncillaryVariableTestMixin):
    # Test for AncillaryVariable creation, with real / lazy data
    def setUp(self):
        self.setupTestArrays(masked=True)

    def test_lazyness_and_dtype_combinations(self):
        for (ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self,
        ):
            data = ancill_var.core_data()
            # Check properties of data.
            if data_lazyness == "real":
                # Real data.
                if ancill_var.dtype == self.data_real.dtype:
                    self.assertArraysShareData(
                        data,
                        self.data_real,
                        "Data values are not the same "
                        "data as the provided array.",
                    )
                    self.assertIsNot(
                        data,
                        self.data_real,
                        "Data array is the same instance as the provided "
                        "array.",
                    )
                else:
                    # the original data values were cast to a test dtype.
                    check_data = self.data_real.astype(ancill_var.dtype)
                    self.assertEqualRealArraysAndDtypes(data, check_data)
            else:
                # Lazy data : the core data may be promoted to float.
                check_data = self.data_lazy.astype(data.dtype)
                self.assertEqualLazyArraysAndDtypes(data, check_data)
                # The realisation type should be correct, though.
                target_dtype = ancill_var.dtype
                self.assertEqual(ancill_var.data.dtype, target_dtype)

    def test_no_masked_data_real(self):
        data = self.no_masked_data_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertEqual(ma.count_masked(data), 0)
        ancill_var = AncillaryVariable(data)
        self.assertFalse(ancill_var.has_lazy_data())
        self.assertTrue(ma.isMaskedArray(ancill_var.data))
        self.assertEqual(ma.count_masked(ancill_var.data), 0)

    def test_no_masked_data_lazy(self):
        data = self.no_masked_data_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertEqual(ma.count_masked(computed), 0)
        ancill_var = AncillaryVariable(data)
        self.assertTrue(ancill_var.has_lazy_data())
        self.assertTrue(ma.isMaskedArray(ancill_var.data))
        self.assertEqual(ma.count_masked(ancill_var.data), 0)

    def test_masked_data_real(self):
        data = self.masked_data_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertTrue(ma.count_masked(data))
        ancill_var = AncillaryVariable(data)
        self.assertFalse(ancill_var.has_lazy_data())
        self.assertTrue(ma.isMaskedArray(ancill_var.data))
        self.assertTrue(ma.count_masked(ancill_var.data))

    def test_masked_data_lazy(self):
        data = self.masked_data_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertTrue(ma.count_masked(computed))
        ancill_var = AncillaryVariable(data)
        self.assertTrue(ancill_var.has_lazy_data())
        self.assertTrue(ma.isMaskedArray(ancill_var.data))
        self.assertTrue(ma.count_masked(ancill_var.data))


class Test_core_data(tests.IrisTest, AncillaryVariableTestMixin):
    # Test for AncillaryVariable.core_data() with various lazy/real data.
    def setUp(self):
        self.setupTestArrays()

    def test_real_data(self):
        ancill_var = AncillaryVariable(self.data_real)
        result = ancill_var.core_data()
        self.assertArraysShareData(
            result,
            self.data_real,
            "core_data() do not share data with the internal array.",
        )

    def test_lazy_data(self):
        ancill_var = AncillaryVariable(self.data_lazy)
        result = ancill_var.core_data()
        self.assertEqualLazyArraysAndDtypes(result, self.data_lazy)

    def test_lazy_points_realise(self):
        ancill_var = AncillaryVariable(self.data_lazy)
        real_data = ancill_var.data
        result = ancill_var.core_data()
        self.assertEqualRealArraysAndDtypes(result, real_data)


class Test_lazy_data(tests.IrisTest, AncillaryVariableTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_core(self):
        ancill_var = AncillaryVariable(self.data_real)
        result = ancill_var.lazy_data()
        self.assertEqualLazyArraysAndDtypes(result, self.data_lazy)

    def test_lazy_core(self):
        ancill_var = AncillaryVariable(self.data_lazy)
        result = ancill_var.lazy_data()
        self.assertIs(result, self.data_lazy)


class Test_has_lazy_data(tests.IrisTest, AncillaryVariableTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_core(self):
        ancill_var = AncillaryVariable(self.data_real)
        result = ancill_var.has_lazy_data()
        self.assertFalse(result)

    def test_lazy_core(self):
        ancill_var = AncillaryVariable(self.data_lazy)
        result = ancill_var.has_lazy_data()
        self.assertTrue(result)

    def test_lazy_core_realise(self):
        ancill_var = AncillaryVariable(self.data_lazy)
        ancill_var.data
        result = ancill_var.has_lazy_data()
        self.assertFalse(result)


class Test__getitem__(tests.IrisTest, AncillaryVariableTestMixin):
    # Test for AncillaryVariable indexing with various types of data.
    def setUp(self):
        self.setupTestArrays()

    def test_partial_slice_data_copy(self):
        parent_ancill_var = AncillaryVariable([1.0, 2.0, 3.0])
        sub_ancill_var = parent_ancill_var[:1]
        values_before_change = sub_ancill_var.data.copy()
        parent_ancill_var.data[:] = -999.9
        self.assertArrayEqual(sub_ancill_var.data, values_before_change)

    def test_full_slice_data_copy(self):
        parent_ancill_var = AncillaryVariable([1.0, 2.0, 3.0])
        sub_ancill_var = parent_ancill_var[:]
        values_before_change = sub_ancill_var.data.copy()
        parent_ancill_var.data[:] = -999.9
        self.assertArrayEqual(sub_ancill_var.data, values_before_change)

    def test_dtypes(self):
        # Index ancillary variables with real+lazy data, and either an int or
        # floating dtype.
        # Check that dtypes remain the same in all cases, taking the dtypes
        # directly from the core data as we have no masking).
        for (main_ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self
        ):

            sub_ancill_var = main_ancill_var[:2, 1]

            ancill_var_dtype = main_ancill_var.dtype
            msg = (
                "Indexing main_ancill_var of dtype {} with {} data changed"
                "dtype of {} to {}."
            )

            sub_data = sub_ancill_var.core_data()
            self.assertEqual(
                sub_data.dtype,
                ancill_var_dtype,
                msg.format(
                    ancill_var_dtype, data_lazyness, "data", sub_data.dtype
                ),
            )

    def test_lazyness(self):
        # Index ancillary variables with real+lazy data, and either an int or
        # floating dtype.
        # Check that lazy data stays lazy and real stays real, in all cases.
        for (main_ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self
        ):

            sub_ancill_var = main_ancill_var[:2, 1]

            msg = (
                "Indexing main_ancill_var of dtype {} with {} data "
                "changed laziness of {} from {!r} to {!r}."
            )
            ancill_var_dtype = main_ancill_var.dtype
            sub_data_lazyness = lazyness_string(sub_ancill_var.core_data())
            self.assertEqual(
                sub_data_lazyness,
                data_lazyness,
                msg.format(
                    ancill_var_dtype,
                    data_lazyness,
                    "data",
                    data_lazyness,
                    sub_data_lazyness,
                ),
            )

    def test_real_data_copies(self):
        # Index ancillary variables with real+lazy data.
        # In all cases, check that any real arrays are copied by the indexing.
        for (main_ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self
        ):

            sub_ancill_var = main_ancill_var[:2, 1]

            msg = (
                "Indexed ancillary variable with {} data "
                "does not have its own separate {} array."
            )
            if data_lazyness == "real":
                main_data = main_ancill_var.core_data()
                sub_data = sub_ancill_var.core_data()
                sub_main_data = main_data[:2, 1]
                self.assertEqualRealArraysAndDtypes(sub_data, sub_main_data)
                self.assertArraysDoNotShareData(
                    sub_data,
                    sub_main_data,
                    msg.format(data_lazyness, "points"),
                )


class Test_copy(tests.IrisTest, AncillaryVariableTestMixin):
    # Test for AncillaryVariable.copy() with various types of data.
    def setUp(self):
        self.setupTestArrays()

    def test_lazyness(self):
        # Copy ancillary variables with real+lazy data, and either an int or
        # floating dtype.
        # Check that lazy data stays lazy and real stays real, in all cases.
        for (main_ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self
        ):

            ancill_var_dtype = main_ancill_var.dtype
            copied_ancill_var = main_ancill_var.copy()

            msg = (
                "Copying main_ancill_var of dtype {} with {} data "
                "changed lazyness of {} from {!r} to {!r}."
            )

            copied_data_lazyness = lazyness_string(
                copied_ancill_var.core_data()
            )
            self.assertEqual(
                copied_data_lazyness,
                data_lazyness,
                msg.format(
                    ancill_var_dtype,
                    data_lazyness,
                    "points",
                    data_lazyness,
                    copied_data_lazyness,
                ),
            )

    def test_realdata_copies(self):
        # Copy ancillary variables with real+lazy data.
        # In all cases, check that any real arrays are copies, not views.
        for (main_ancill_var, data_lazyness) in data_all_dtypes_and_lazynesses(
            self
        ):

            copied_ancill_var = main_ancill_var.copy()

            msg = (
                "Copied ancillary variable with {} data "
                "does not have its own separate {} array."
            )

            if data_lazyness == "real":
                main_data = main_ancill_var.core_data()
                copied_data = copied_ancill_var.core_data()
                self.assertEqualRealArraysAndDtypes(main_data, copied_data)
                self.assertArraysDoNotShareData(
                    main_data, copied_data, msg.format(data_lazyness, "points")
                )


class Test_data__getter(tests.IrisTest, AncillaryVariableTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_mutable_real_data(self):
        # Check that ancill_var.data returns a modifiable array, and changes
        # to it are reflected to the ancillary_var.
        data = np.array([1.0, 2.0, 3.0, 4.0])
        ancill_var = AncillaryVariable(data)
        initial_values = data.copy()
        ancill_var.data[1:2] += 33.1
        result = ancill_var.data
        self.assertFalse(np.all(result == initial_values))

    def test_real_data(self):
        # Getting real data does not change or copy them.
        ancill_var = AncillaryVariable(self.data_real)
        result = ancill_var.data
        self.assertArraysShareData(
            result,
            self.data_real,
            "Data values do not share data with the provided array.",
        )

    def test_lazy_data(self):
        # Getting lazy data realises them.
        ancill_var = AncillaryVariable(self.data_lazy)
        self.assertTrue(ancill_var.has_lazy_data())
        result = ancill_var.data
        self.assertFalse(ancill_var.has_lazy_data())
        self.assertEqualRealArraysAndDtypes(result, self.data_real)


class Test_data__setter(tests.IrisTest, AncillaryVariableTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_set_real(self):
        # Setting new real data does not make a copy.
        ancill_var = AncillaryVariable(self.data_real)
        new_data = self.data_real + 102.3
        ancill_var.data = new_data
        result = ancill_var.core_data()
        self.assertArraysShareData(
            result,
            new_data,
            "Data values do not share data with the assigned array.",
        )

    def test_fail_bad_shape(self):
        # Setting real data requires matching shape.
        ancill_var = AncillaryVariable([1.0, 2.0])
        msg = r"Require data with shape \(2,\), got \(3,\)"
        with self.assertRaisesRegex(ValueError, msg):
            ancill_var.data = np.array([1.0, 2.0, 3.0])

    def test_real_set_lazy(self):
        # Setting new lazy data does not make a copy.
        ancill_var = AncillaryVariable(self.data_real)
        new_data = self.data_lazy + 102.3
        ancill_var.data = new_data
        result = ancill_var.core_data()
        self.assertEqualLazyArraysAndDtypes(result, new_data)


class Test__str__(tests.IrisTest):
    def test_non_time_values(self):
        ancillary_var = AncillaryVariable(
            np.array([2, 5, 9]),
            standard_name="height",
            long_name="height of detector",
            var_name="height",
            units="m",
            attributes={"notes": "Measured from sea level"},
        )
        expected = (
            "AncillaryVariable(array([2, 5, 9]), "
            "standard_name='height', units=Unit('m'), "
            "long_name='height of detector', var_name='height', "
            "attributes={'notes': 'Measured from sea level'})"
        )
        self.assertEqual(expected, ancillary_var.__str__())

    def test_time_values(self):
        ancillary_var = AncillaryVariable(
            np.array([2, 5, 9]),
            units="hours since 1970-01-01 01:00",
            long_name="time of previous valid detection",
        )
        expected = (
            "AncillaryVariable([1970-01-01 03:00:00, "
            "1970-01-01 06:00:00, 1970-01-01 10:00:00], "
            "standard_name=None, calendar='gregorian', "
            "long_name='time of previous valid detection')"
        )
        self.assertEqual(expected, ancillary_var.__str__())


class Test__repr__(tests.IrisTest):
    def test_non_time_values(self):
        ancillary_var = AncillaryVariable(
            np.array([2, 5, 9]),
            standard_name="height",
            long_name="height of detector",
            var_name="height",
            units="m",
            attributes={"notes": "Measured from sea level"},
        )
        expected = (
            "AncillaryVariable(array([2, 5, 9]), "
            "standard_name='height', units=Unit('m'), "
            "long_name='height of detector', var_name='height', "
            "attributes={'notes': 'Measured from sea level'})"
        )
        self.assertEqual(expected, ancillary_var.__repr__())

    def test_time_values(self):
        ancillary_var = AncillaryVariable(
            np.array([2, 5, 9]),
            units="hours since 1970-01-01 01:00",
            long_name="time of previous valid detection",
        )
        expected = (
            "AncillaryVariable(array([2, 5, 9]), standard_name=None, "
            "units=Unit('hours since 1970-01-01 01:00', "
            "calendar='gregorian'), "
            "long_name='time of previous valid detection')"
        )
        self.assertEqual(expected, ancillary_var.__repr__())


class Test___binary_operator__(tests.IrisTest, AncillaryVariableTestMixin):
    # Test maths operations on on real+lazy data.
    def setUp(self):
        self.setupTestArrays()

        self.real_ancill_var = AncillaryVariable(self.data_real)
        self.lazy_ancill_var = AncillaryVariable(self.data_lazy)

        self.test_combinations = [
            (self.real_ancill_var, self.data_real, "real"),
            (self.lazy_ancill_var, self.data_lazy, "lazy"),
        ]

    def _check(self, result_ancill_var, expected_data, lazyness):
        # Test each operation on
        data = result_ancill_var.core_data()
        if lazyness == "real":
            self.assertEqualRealArraysAndDtypes(expected_data, data)
        else:
            self.assertEqualLazyArraysAndDtypes(expected_data, data)

    def test_add(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = ancill_var + 10
            expected_data = orig_data + 10
            self._check(result, expected_data, data_lazyness)

    def test_add_inplace(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            ancill_var += 10
            expected_data = orig_data + 10
            self._check(ancill_var, expected_data, data_lazyness)

    def test_right_add(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = 10 + ancill_var
            expected_data = 10 + orig_data
            self._check(result, expected_data, data_lazyness)

    def test_subtract(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = ancill_var - 10
            expected_data = orig_data - 10
            self._check(result, expected_data, data_lazyness)

    def test_subtract_inplace(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            ancill_var -= 10
            expected_data = orig_data - 10
            self._check(ancill_var, expected_data, data_lazyness)

    def test_right_subtract(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = 10 - ancill_var
            expected_data = 10 - orig_data
            self._check(result, expected_data, data_lazyness)

    def test_multiply(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = ancill_var * 10
            expected_data = orig_data * 10
            self._check(result, expected_data, data_lazyness)

    def test_multiply_inplace(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            ancill_var *= 10
            expected_data = orig_data * 10
            self._check(ancill_var, expected_data, data_lazyness)

    def test_right_multiply(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = 10 * ancill_var
            expected_data = 10 * orig_data
            self._check(result, expected_data, data_lazyness)

    def test_divide(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = ancill_var / 10
            expected_data = orig_data / 10
            self._check(result, expected_data, data_lazyness)

    def test_divide_inplace(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            ancill_var /= 10
            expected_data = orig_data / 10
            self._check(ancill_var, expected_data, data_lazyness)

    def test_right_divide(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = 10 / ancill_var
            expected_data = 10 / orig_data
            self._check(result, expected_data, data_lazyness)

    def test_negative(self):
        for (ancill_var, orig_data, data_lazyness) in self.test_combinations:
            result = -ancill_var
            expected_data = -orig_data
            self._check(result, expected_data, data_lazyness)


class Test_has_bounds(tests.IrisTest):
    def test(self):
        ancillary_var = AncillaryVariable(np.array([2, 9, 5]))
        self.assertFalse(ancillary_var.has_bounds())


class Test_convert_units(tests.IrisTest):
    def test_preserves_lazy(self):
        test_data = np.array([[11.1, 12.2, 13.3], [21.4, 22.5, 23.6]])
        lazy_data = as_lazy_data(test_data)
        ancill_var = AncillaryVariable(data=lazy_data, units="m")
        ancill_var.convert_units("ft")
        self.assertTrue(ancill_var.has_lazy_data())
        test_data_ft = Unit("m").convert(test_data, "ft")
        self.assertArrayAllClose(ancill_var.data, test_data_ft)


class Test_is_compatible(tests.IrisTest):
    def setUp(self):
        self.ancill_var = AncillaryVariable(
            [1.0, 8.0, 22.0], standard_name="number_of_observations", units="1"
        )
        self.modified_ancill_var = self.ancill_var.copy()

    def test_not_compatible_diff_name(self):
        # Different name() - not compatible
        self.modified_ancill_var.rename("air_temperature")
        self.assertFalse(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_not_compatible_diff_units(self):
        # Different units- not compatible
        self.modified_ancill_var.units = "m"
        self.assertFalse(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_not_compatible_diff_common_attrs(self):
        # Different common attributes - not compatible.
        self.ancill_var.attributes["source"] = "A"
        self.modified_ancill_var.attributes["source"] = "B"
        self.assertFalse(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_compatible_diff_data(self):
        # Different data values - compatible.
        self.modified_ancill_var.data = [10.0, 20.0, 100.0]
        self.assertTrue(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_compatible_diff_var_name(self):
        # Different var_name (but same name()) - compatible.
        self.modified_ancill_var.var_name = "obs_num"
        self.assertTrue(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_compatible_diff_non_common_attributes(self):
        # Different non-common attributes - compatible.
        self.ancill_var.attributes["source"] = "A"
        self.modified_ancill_var.attributes["origin"] = "B"
        self.assertTrue(
            self.ancill_var.is_compatible(self.modified_ancill_var)
        )

    def test_compatible_ignore_common_attribute(self):
        # ignore different common attributes - compatible.
        self.ancill_var.attributes["source"] = "A"
        self.modified_ancill_var.attributes["source"] = "B"
        self.assertTrue(
            self.ancill_var.is_compatible(
                self.modified_ancill_var, ignore="source"
            )
        )


class TestEquality(tests.IrisTest):
    def test_nanpoints_eq_self(self):
        av1 = AncillaryVariable([1.0, np.nan, 2.0])
        self.assertEqual(av1, av1)

    def test_nanpoints_eq_copy(self):
        av1 = AncillaryVariable([1.0, np.nan, 2.0])
        av2 = av1.copy()
        self.assertEqual(av1, av2)


class Test_cube_dims(tests.IrisTest):
    def test(self):
        # Check that "coord.cube_dims(cube)" calls "cube.coord_dims(coord)".
        mock_dims_result = mock.sentinel.AV_DIMS
        mock_dims_call = mock.Mock(return_value=mock_dims_result)
        mock_cube = mock.Mock(Cube, ancillary_variable_dims=mock_dims_call)
        test_var = AncillaryVariable([1], long_name="test_name")

        result = test_var.cube_dims(mock_cube)
        self.assertEqual(result, mock_dims_result)
        self.assertEqual(mock_dims_call.call_args_list, [mock.call(test_var)])


if __name__ == "__main__":
    tests.main()



import matplotlib.pyplot as plt
import iris
import iris.quickplot as qplt
import iris.plot as iplt


fname = iris.sample_data_path("air_temp.pp")
temperature_cube = iris.load_cube(fname)

# Get the Purples "Brewer" palette.
brewer_cmap = plt.get_cmap("brewer_Purples_09")

# Draw the contours, with n-levels set for the map colours (9).
# NOTE: needed as the map is non-interpolated, but matplotlib does not provide
# any special behaviour for these.
qplt.contourf(temperature_cube, brewer_cmap.N, cmap=brewer_cmap)

# Add a citation to the plot.
iplt.citation(iris.plot.BREWER_CITE)

# Add coastlines to the map created by contourf.
plt.gca().coastlines()

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests to check the validity of replacing
"iris.analysis._interpolate.regrid`('nearest')" with
"iris.cube.Cube.regrid(scheme=iris.analysis.Nearest())".

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.analysis import Nearest
from iris.cube import Cube
from iris.coords import DimCoord


def grid_cube(xx, yy, data=None):
    nx, ny = len(xx), len(yy)
    if data is not None:
        data = np.array(data).reshape((ny, nx))
    else:
        data = np.zeros((ny, nx))
    cube = Cube(data)
    y_coord = DimCoord(yy, standard_name="latitude", units="degrees")
    x_coord = DimCoord(xx, standard_name="longitude", units="degrees")
    cube.add_dim_coord(y_coord, 0)
    cube.add_dim_coord(x_coord, 1)
    return cube


ENABLE_DEBUG_OUTPUT = False


def _debug_data(cube, test_id):
    if ENABLE_DEBUG_OUTPUT:
        print
        data = cube.data
        print("CUBE: {}".format(test_id))
        print("  x={!r}".format(cube.coord("longitude").points))
        print("  y={!r}".format(cube.coord("latitude").points))
        print("data[{}]:".format(type(data)))
        print(repr(data))


class MixinCheckingCode:
    def test_basic(self):
        src_x = [30.0, 40.0, 50.0]
        dst_x = [32.0, 42.0]
        src_y = [-10.0, 0.0, 10.0]
        dst_y = [-8.0, 2.0]
        data = [[3.0, 4.0, 5.0], [23.0, 24.0, 25.0], [43.0, 44.0, 45.0]]
        expected_result = [[3.0, 4.0], [23.0, 24.0]]
        src_cube = grid_cube(src_x, src_y, data)
        _debug_data(src_cube, "basic SOURCE")
        dst_cube = grid_cube(dst_x, dst_y)
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "basic RESULT")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_src_extrapolation(self):
        src_x = [30.0, 40.0, 50.0]
        dst_x = [0.0, 29.0, 39.0]
        src_y = [-10.0, 0.0, 10.0]
        dst_y = [-50.0, -9.0, -1.0]
        data = [[3.0, 4.0, 5.0], [23.0, 24.0, 25.0], [43.0, 44.0, 45.0]]
        expected_result = [
            [3.0, 3.0, 4.0],
            [3.0, 3.0, 4.0],
            [23.0, 23.0, 24.0],
        ]
        src_cube = grid_cube(src_x, src_y, data)
        _debug_data(src_cube, "extrapolate SOURCE")
        dst_cube = grid_cube(dst_x, dst_y)
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "extrapolate RESULT")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_exact_matching_points(self):
        src_x = [10.0, 20.0, 30.0]
        src_y = [10.0, 20.0, 30.0]
        dst_x = [14.9, 15.1, 20.0, 24.9, 25.1]
        dst_y = [14.9, 15.1, 20.0, 24.9, 25.1]
        data = [[3.0, 4.0, 5.0], [23.0, 24.0, 25.0], [43.0, 44.0, 45.0]]
        expected_result = [
            [3.0, 4.0, 4.0, 4.0, 5.0],
            [23.0, 24.0, 24.0, 24.0, 25.0],
            [23.0, 24.0, 24.0, 24.0, 25.0],
            [23.0, 24.0, 24.0, 24.0, 25.0],
            [43.0, 44.0, 44.0, 44.0, 45.0],
        ]
        src_cube = grid_cube(src_x, src_y, data)
        _debug_data(src_cube, "matching SOURCE")
        dst_cube = grid_cube(dst_x, dst_y)
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "matching RESULt")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_source_mask(self):
        src_x = [40.0, 50.0, 60.0]
        src_y = [40.0, 50.0, 60.0]
        dst_x = [44.99, 45.01, 48.0, 50.0, 52.0, 54.99, 55.01]
        dst_y = [44.99, 45.01, 48.0, 50.0, 52.0, 54.99, 55.01]
        data = np.ma.masked_equal(
            [[3.0, 4.0, 5.0], [23.0, 999, 25.0], [43.0, 44.0, 45.0]], 999
        )
        expected_result = np.ma.masked_equal(
            [
                [3.0, 4.0, 4.0, 4.0, 4.0, 4.0, 5.0],
                [23.0, 999, 999, 999, 999, 999, 25.0],
                [23.0, 999, 999, 999, 999, 999, 25.0],
                [23.0, 999, 999, 999, 999, 999, 25.0],
                [23.0, 999, 999, 999, 999, 999, 25.0],
                [23.0, 999, 999, 999, 999, 999, 25.0],
                [43.0, 44.0, 44.0, 44.0, 44.0, 44.0, 45.0],
            ],
            999,
        )
        src_cube = grid_cube(src_x, src_y, data)
        src_cube.data = np.ma.masked_array(src_cube.data)
        src_cube.data[1, 1] = np.ma.masked
        _debug_data(src_cube, "masked SOURCE")
        dst_cube = grid_cube(dst_x, dst_y)
        result_cube = self.regrid(
            src_cube, dst_cube, translate_nans_to_mask=True
        )
        _debug_data(result_cube, "masked RESULT")
        self.assertMaskedArrayEqual(result_cube.data, expected_result)

    def test_wrapping_non_circular(self):
        src_x = [-10.0, 0.0, 10.0]
        dst_x = [-360.0, -170.0, -1.0, 1.0, 50.0, 170.0, 352.0, 720.0]
        src_y = [0.0, 10.0]
        dst_y = [0.0, 10.0]
        data = [[3.0, 4.0, 5.0], [3.0, 4.0, 5.0]]
        src_cube = grid_cube(src_x, src_y, data)
        dst_cube = grid_cube(dst_x, dst_y)
        expected_result = [
            [4.0, 3.0, 4.0, 4.0, 5.0, 5.0, 3.0, 4.0],
            [4.0, 3.0, 4.0, 4.0, 5.0, 5.0, 3.0, 4.0],
        ]
        _debug_data(src_cube, "noncircular SOURCE")
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "noncircular RESULT")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_wrapping_circular(self):
        # When x-coord is "circular", the above distinction does not apply :
        # results are the same for both calculations.
        src_x = [-10.0, 0.0, 10.0]
        dst_x = [-360.0, -170.0, -1.0, 1.0, 50.0, 170.0, 352.0, 720.0]
        src_y = [0.0, 10.0]
        dst_y = [0.0, 10.0]
        data = [[3.0, 4.0, 5.0], [3.0, 4.0, 5.0]]
        src_cube = grid_cube(src_x, src_y, data)
        dst_cube = grid_cube(dst_x, dst_y)
        src_cube.coord("longitude").circular = True
        expected_result = [
            [4.0, 3.0, 4.0, 4.0, 5.0, 5.0, 3.0, 4.0],
            [4.0, 3.0, 4.0, 4.0, 5.0, 5.0, 3.0, 4.0],
        ]
        _debug_data(src_cube, "circular SOURCE")
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "circular RESULT")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_wrapping_non_angular(self):
        src_x = [-10.0, 0.0, 10.0]
        dst_x = [-360.0, -170.0, -1.0, 1.0, 50.0, 170.0, 352.0, 720.0]
        src_y = [0.0, 10.0]
        dst_y = [0.0, 10.0]
        data = [[3.0, 4.0, 5.0], [3.0, 4.0, 5.0]]
        src_cube = grid_cube(src_x, src_y, data)
        dst_cube = grid_cube(dst_x, dst_y)
        for co_name in ("longitude", "latitude"):
            for cube in (src_cube, dst_cube):
                coord = cube.coord(co_name)
                coord.coord_system = None
                coord.convert_units("1")
        # interpolate.regrid --> Wrapping-free results (non-circular).
        expected_result = [
            [3.0, 3.0, 4.0, 4.0, 5.0, 5.0, 5.0, 5.0],
            [3.0, 3.0, 4.0, 4.0, 5.0, 5.0, 5.0, 5.0],
        ]
        _debug_data(src_cube, "non-angle-lons SOURCE")
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "non-angle-lons RESULT")
        self.assertArrayAllClose(result_cube.data, expected_result)

    def test_source_nan(self):
        src_x = [40.0, 50.0, 60.0]
        src_y = [40.0, 50.0, 60.0]
        dst_x = [44.99, 45.01, 48.0, 50.0, 52.0, 54.99, 55.01]
        dst_y = [44.99, 45.01, 48.0, 50.0, 52.0, 54.99, 55.01]
        nan = np.nan
        data = [[3.0, 4.0, 5.0], [23.0, nan, 25.0], [43.0, 44.0, 45.0]]
        expected_result = [
            [3.0, 4.0, 4.0, 4.0, 4.0, 4.0, 5.0],
            [23.0, nan, nan, nan, nan, nan, 25.0],
            [23.0, nan, nan, nan, nan, nan, 25.0],
            [23.0, nan, nan, nan, nan, nan, 25.0],
            [23.0, nan, nan, nan, nan, nan, 25.0],
            [23.0, nan, nan, nan, nan, nan, 25.0],
            [43.0, 44.0, 44.0, 44.0, 44.0, 44.0, 45.0],
        ]
        src_cube = grid_cube(src_x, src_y, data)
        _debug_data(src_cube, "nan SOURCE")
        dst_cube = grid_cube(dst_x, dst_y)
        result_cube = self.regrid(src_cube, dst_cube)
        _debug_data(result_cube, "nan RESULT")
        self.assertArrayEqual(result_cube.data, expected_result)


class TestCubeRegridNearest(MixinCheckingCode, tests.IrisTest):
    scheme = Nearest(extrapolation_mode="extrapolate")

    def regrid(self, src_cube, dst_cube, **kwargs):
        return src_cube.regrid(dst_cube, scheme=self.scheme)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Exceptions specific to the Iris package.

"""


class IrisError(Exception):
    """Base class for errors in the Iris package."""

    pass


class CoordinateCollapseError(IrisError):
    """Raised when a requested coordinate cannot be collapsed."""

    pass


class CoordinateNotFoundError(KeyError):
    """Raised when a search yields no coordinates."""

    pass


class CellMeasureNotFoundError(KeyError):
    """Raised when a search yields no cell measures."""

    pass


class AncillaryVariableNotFoundError(KeyError):
    """Raised when a search yields no ancillary variables."""

    pass


class CoordinateMultiDimError(ValueError):
    """Raised when a routine doesn't support multi-dimensional coordinates."""

    def __init__(self, msg):
        # N.B. deferred import to avoid a circular import dependency.
        import iris.coords

        if isinstance(msg, iris.coords.Coord):
            fmt = "Multi-dimensional coordinate not supported: '%s'"
            msg = fmt % msg.name()
        ValueError.__init__(self, msg)


class CoordinateNotRegularError(ValueError):
    """Raised when a coordinate is unexpectedly irregular."""

    pass


class InvalidCubeError(IrisError):
    """Raised when a Cube validation check fails."""

    pass


class ConstraintMismatchError(IrisError):
    """
    Raised when a constraint operation has failed to find the correct number
    of results.

    """

    pass


class NotYetImplementedError(IrisError):
    """
    Raised by missing functionality.

    Different meaning to NotImplementedError, which is for abstract methods.

    """

    pass


class TranslationError(IrisError):
    """Raised when Iris is unable to translate format-specific codes."""

    pass


class IgnoreCubeException(IrisError):
    """
    Raised from a callback function when a cube should be ignored on load.

    """

    pass


class ConcatenateError(IrisError):
    """
    Raised when concatenate is expected to produce a single cube, but fails to
    do so.

    """

    def __init__(self, differences):
        """
        Creates a ConcatenateError with a list of textual descriptions of
        the differences which prevented a concatenate.

        Args:

        * differences:
            The list of strings which describe the differences.

        """
        self.differences = differences

    def __str__(self):
        return "\n  ".join(
            ["failed to concatenate into a single cube."]
            + list(self.differences)
        )


class MergeError(IrisError):
    """
    Raised when merge is expected to produce a single cube, but fails to
    do so.

    """

    def __init__(self, differences):
        """
        Creates a MergeError with a list of textual descriptions of
        the differences which prevented a merge.

        Args:

        * differences:
            The list of strings which describe the differences.

        """
        self.differences = differences

    def __str__(self):
        return "\n  ".join(
            ["failed to merge into a single cube."] + list(self.differences)
        )


class DuplicateDataError(MergeError):
    """Raised when merging two or more cubes that have identical metadata."""

    def __init__(self, msg):
        self.differences = [msg]


class LazyAggregatorError(Exception):
    pass


class UnitConversionError(IrisError):
    """Raised when Iris is unable to convert a unit."""

    pass



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.


# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import cf_units
import numpy as np
import numpy.ma as ma

import iris
import iris.analysis.cartography
import iris.analysis.maths
import iris.coord_systems
import iris.coords
import iris.cube
import iris.tests.stock


class TestAnalysisCubeCoordComparison(tests.IrisTest):
    def assertComparisonDict(self, comparison_dict, reference_filename):
        string = ""
        for key in sorted(comparison_dict):
            coord_groups = comparison_dict[key]
            string += "%40s  " % key
            names = [
                [
                    coord.name() if coord is not None else "None"
                    for coord in coords
                ]
                for coords in coord_groups
            ]
            string += str(sorted(names))
            string += "\n"
        self.assertString(string, reference_filename)

    def test_coord_comparison(self):
        cube1 = iris.cube.Cube(np.zeros((41, 41)))
        lonlat_cs = iris.coord_systems.GeogCS(6371229)
        lon_points1 = -180 + 4.5 * np.arange(41, dtype=np.float32)
        lat_points = -90 + 4.5 * np.arange(41, dtype=np.float32)
        cube1.add_dim_coord(
            iris.coords.DimCoord(
                lon_points1,
                "longitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            0,
        )
        cube1.add_dim_coord(
            iris.coords.DimCoord(
                lat_points, "latitude", units="degrees", coord_system=lonlat_cs
            ),
            1,
        )
        cube1.add_aux_coord(iris.coords.AuxCoord(0, long_name="z"))
        cube1.add_aux_coord(
            iris.coords.AuxCoord(["foobar"], long_name="f", units="no_unit")
        )

        cube2 = iris.cube.Cube(np.zeros((41, 41, 5)))
        lonlat_cs = iris.coord_systems.GeogCS(6371229)
        lon_points2 = -160 + 4.5 * np.arange(41, dtype=np.float32)
        cube2.add_dim_coord(
            iris.coords.DimCoord(
                lon_points2,
                "longitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            0,
        )
        cube2.add_dim_coord(
            iris.coords.DimCoord(
                lat_points, "latitude", units="degrees", coord_system=lonlat_cs
            ),
            1,
        )
        cube2.add_dim_coord(
            iris.coords.DimCoord([5, 7, 9, 11, 13], long_name="z"), 2
        )

        cube3 = cube1.copy()
        lon = cube3.coord("longitude")
        lat = cube3.coord("latitude")
        cube3.remove_coord(lon)
        cube3.remove_coord(lat)
        cube3.add_dim_coord(lon, 1)
        cube3.add_dim_coord(lat, 0)
        cube3.coord("z").points = [20]

        cube4 = cube2.copy()
        lon = cube4.coord("longitude")
        lat = cube4.coord("latitude")
        cube4.remove_coord(lon)
        cube4.remove_coord(lat)
        cube4.add_dim_coord(lon, 1)
        cube4.add_dim_coord(lat, 0)

        # Test when coords are the same object
        lon = cube1.coord("longitude")
        lat = cube1.coord("latitude")
        cube5 = iris.cube.Cube(np.zeros((41, 41)))
        cube5.add_dim_coord(lon, 0)
        cube5.add_dim_coord(lat, 1)

        coord_comparison = iris.analysis._dimensional_metadata_comparison

        self.assertComparisonDict(
            coord_comparison(cube1, cube1),
            ("analysis", "coord_comparison", "cube1_cube1.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube1, cube2),
            ("analysis", "coord_comparison", "cube1_cube2.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube1, cube3),
            ("analysis", "coord_comparison", "cube1_cube3.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube1, cube4),
            ("analysis", "coord_comparison", "cube1_cube4.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube1, cube5),
            ("analysis", "coord_comparison", "cube1_cube5.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube2, cube3),
            ("analysis", "coord_comparison", "cube2_cube3.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube2, cube4),
            ("analysis", "coord_comparison", "cube2_cube4.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube2, cube5),
            ("analysis", "coord_comparison", "cube2_cube5.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube3, cube4),
            ("analysis", "coord_comparison", "cube3_cube4.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube3, cube5),
            ("analysis", "coord_comparison", "cube3_cube5.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube4, cube5),
            ("analysis", "coord_comparison", "cube4_cube5.txt"),
        )

        self.assertComparisonDict(
            coord_comparison(cube1, cube1, cube1),
            ("analysis", "coord_comparison", "cube1_cube1_cube1.txt"),
        )
        self.assertComparisonDict(
            coord_comparison(cube1, cube2, cube1),
            ("analysis", "coord_comparison", "cube1_cube2_cube1.txt"),
        )

        # get a coord comparison result and check that we are getting back what was expected
        coord_group = coord_comparison(cube1, cube2)["grouped_coords"][0]
        self.assertIsInstance(coord_group, iris.analysis._CoordGroup)
        self.assertIsInstance(list(coord_group)[0], iris.coords.Coord)


class TestAnalysisWeights(tests.IrisTest):
    def test_weighted_mean_little(self):
        data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float32)
        weights = np.array([[9, 8, 7], [6, 5, 4], [3, 2, 1]], dtype=np.float32)

        cube = iris.cube.Cube(data, long_name="test_data", units="1")
        hcs = iris.coord_systems.GeogCS(6371229)
        lat_coord = iris.coords.DimCoord(
            np.array([1, 2, 3], dtype=np.float32),
            long_name="lat",
            units="1",
            coord_system=hcs,
        )
        lon_coord = iris.coords.DimCoord(
            np.array([1, 2, 3], dtype=np.float32),
            long_name="lon",
            units="1",
            coord_system=hcs,
        )
        cube.add_dim_coord(lat_coord, 0)
        cube.add_dim_coord(lon_coord, 1)
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(3, dtype=np.float32), long_name="dummy", units=1
            ),
            1,
        )
        self.assertCML(cube, ("analysis", "weighted_mean_source.cml"))

        a = cube.collapsed("lat", iris.analysis.MEAN, weights=weights)
        # np.ma.average doesn't apply type promotion rules in some versions,
        # and instead makes the result type float64. To ignore that case we
        # fix up the dtype here if it is promotable from float32. We still want
        # to catch cases where there is a loss of precision however.
        if a.dtype > np.float32:
            cast_data = a.data.astype(np.float32)
            a.data = cast_data
        self.assertCMLApproxData(a, ("analysis", "weighted_mean_lat.cml"))

        b = cube.collapsed(lon_coord, iris.analysis.MEAN, weights=weights)
        if b.dtype > np.float32:
            cast_data = b.data.astype(np.float32)
            b.data = cast_data
        b.data = np.asarray(b.data)
        self.assertCMLApproxData(b, ("analysis", "weighted_mean_lon.cml"))
        self.assertEqual(b.coord("dummy").shape, (1,))

        # test collapsing multiple coordinates (and the fact that one of the coordinates isn't the same coordinate instance as on the cube)
        c = cube.collapsed(
            [lat_coord[:], lon_coord], iris.analysis.MEAN, weights=weights
        )
        if c.dtype > np.float32:
            cast_data = c.data.astype(np.float32)
            c.data = cast_data
        self.assertCMLApproxData(c, ("analysis", "weighted_mean_latlon.cml"))
        self.assertEqual(c.coord("dummy").shape, (1,))

        # Check new coord bounds - made from points
        self.assertArrayEqual(c.coord("lat").bounds, [[1, 3]])

        # Check new coord bounds - made from bounds
        cube.coord("lat").bounds = [[0.5, 1.5], [1.5, 2.5], [2.5, 3.5]]
        c = cube.collapsed(["lat", "lon"], iris.analysis.MEAN, weights=weights)
        self.assertArrayEqual(c.coord("lat").bounds, [[0.5, 3.5]])
        cube.coord("lat").bounds = None

        # Check there was no residual change
        self.assertCML(cube, ("analysis", "weighted_mean_source.cml"))

    @tests.skip_data
    def test_weighted_mean(self):
        # compare with pp_area_avg - which collapses both lat and lon
        #
        #     pp = ppa('/data/local/dataZoo/PP/simple_pp/global.pp', 0)
        #     print, pp_area(pp, /box)
        #     print, pp_area_avg(pp, /box)  #287.927
        #     ;gives an answer of 287.927
        #
        e = iris.tests.stock.simple_pp()
        self.assertCML(e, ("analysis", "weighted_mean_original.cml"))
        e.coord("latitude").guess_bounds()
        e.coord("longitude").guess_bounds()
        area_weights = iris.analysis.cartography.area_weights(e)
        e.coord("latitude").bounds = None
        e.coord("longitude").bounds = None
        f, collapsed_area_weights = e.collapsed(
            "latitude", iris.analysis.MEAN, weights=area_weights, returned=True
        )
        g = f.collapsed(
            "longitude", iris.analysis.MEAN, weights=collapsed_area_weights
        )
        # check it's a 0d, scalar cube
        self.assertEqual(g.shape, ())
        # check the value - pp_area_avg's result of 287.927 differs by factor of 1.00002959
        np.testing.assert_approx_equal(g.data, 287.935, significant=5)

        # check we get summed weights even if we don't give any
        h, summed_weights = e.collapsed(
            "latitude", iris.analysis.MEAN, returned=True
        )
        assert summed_weights is not None

        # Check there was no residual change
        e.coord("latitude").bounds = None
        e.coord("longitude").bounds = None
        self.assertCML(e, ("analysis", "weighted_mean_original.cml"))

        # Test collapsing of missing coord
        self.assertRaises(
            iris.exceptions.CoordinateNotFoundError,
            e.collapsed,
            "platitude",
            iris.analysis.MEAN,
        )

        # Test collpasing of non data coord
        self.assertRaises(
            iris.exceptions.CoordinateCollapseError,
            e.collapsed,
            "pressure",
            iris.analysis.MEAN,
        )


@tests.skip_data
class TestAnalysisBasic(tests.IrisTest):
    def setUp(self):
        file = tests.get_data_path(("PP", "aPProt1", "rotatedMHtimecube.pp"))
        cubes = iris.load(file)
        self.cube = cubes[0]
        self.assertCML(self.cube, ("analysis", "original.cml"))

    def _common(
        self,
        name,
        aggregate,
        original_name="original_common.cml",
        *args,
        **kwargs,
    ):
        self.cube.data = self.cube.data.astype(np.float64)

        self.assertCML(self.cube, ("analysis", original_name))

        a = self.cube.collapsed("grid_latitude", aggregate)
        self.assertCMLApproxData(
            a, ("analysis", "%s_latitude.cml" % name), *args, **kwargs
        )

        b = a.collapsed("grid_longitude", aggregate)
        self.assertCMLApproxData(
            b,
            ("analysis", "%s_latitude_longitude.cml" % name),
            *args,
            **kwargs,
        )

        c = self.cube.collapsed(["grid_latitude", "grid_longitude"], aggregate)
        self.assertCMLApproxData(
            c,
            ("analysis", "%s_latitude_longitude_1call.cml" % name),
            *args,
            **kwargs,
        )

        # Check there was no residual change
        self.assertCML(self.cube, ("analysis", original_name))

    def test_mean(self):
        self._common("mean", iris.analysis.MEAN, rtol=1e-05)

    def test_std_dev(self):
        # as the numbers are so high, trim off some trailing digits & compare to 0dp
        self._common("std_dev", iris.analysis.STD_DEV, rtol=1e-05)

    def test_hmean(self):
        # harmonic mean requires data > 0
        self.cube.data *= self.cube.data
        self._common(
            "hmean", iris.analysis.HMEAN, "original_hmean.cml", rtol=1e-05
        )

    def test_gmean(self):
        self._common("gmean", iris.analysis.GMEAN, rtol=1e-05)

    def test_variance(self):
        # as the numbers are so high, trim off some trailing digits & compare to 0dp
        self._common("variance", iris.analysis.VARIANCE, rtol=1e-05)

    def test_median(self):
        self._common("median", iris.analysis.MEDIAN)

    def test_sum(self):
        # as the numbers are so high, trim off some trailing digits & compare to 0dp
        self._common("sum", iris.analysis.SUM, rtol=1e-05)

    def test_max(self):
        self._common("max", iris.analysis.MAX)

    def test_min(self):
        self._common("min", iris.analysis.MIN)

    def test_rms(self):
        self._common("rms", iris.analysis.RMS)

    def test_duplicate_coords(self):
        self.assertRaises(ValueError, tests.stock.track_1d, duplicate_x=True)


class TestMissingData(tests.IrisTest):
    def setUp(self):
        self.cube_with_nan = tests.stock.simple_2d()

        data = self.cube_with_nan.data.astype(np.float32)
        self.cube_with_nan.data = data.copy()
        self.cube_with_nan.data[1, 0] = np.nan
        self.cube_with_nan.data[2, 2] = np.nan
        self.cube_with_nan.data[2, 3] = np.nan

        self.cube_with_mask = tests.stock.simple_2d()
        self.cube_with_mask.data = ma.array(
            self.cube_with_nan.data, mask=np.isnan(self.cube_with_nan.data)
        )

    def test_max(self):
        cube = self.cube_with_nan.collapsed("foo", iris.analysis.MAX)
        np.testing.assert_array_equal(cube.data, np.array([3, np.nan, np.nan]))

        cube = self.cube_with_mask.collapsed("foo", iris.analysis.MAX)
        np.testing.assert_array_equal(cube.data, np.array([3, 7, 9]))

    def test_min(self):
        cube = self.cube_with_nan.collapsed("foo", iris.analysis.MIN)
        np.testing.assert_array_equal(cube.data, np.array([0, np.nan, np.nan]))

        cube = self.cube_with_mask.collapsed("foo", iris.analysis.MIN)
        np.testing.assert_array_equal(cube.data, np.array([0, 5, 8]))

    def test_sum(self):
        cube = self.cube_with_nan.collapsed("foo", iris.analysis.SUM)
        np.testing.assert_array_equal(cube.data, np.array([6, np.nan, np.nan]))

        cube = self.cube_with_mask.collapsed("foo", iris.analysis.SUM)
        np.testing.assert_array_equal(cube.data, np.array([6, 18, 17]))


class TestAuxCoordCollapse(tests.IrisTest):
    def setUp(self):
        self.cube_with_aux_coord = tests.stock.simple_4d_with_hybrid_height()

        # Guess bounds to get the weights
        self.cube_with_aux_coord.coord("grid_latitude").guess_bounds()
        self.cube_with_aux_coord.coord("grid_longitude").guess_bounds()

    def test_max(self):
        cube = self.cube_with_aux_coord.collapsed(
            "grid_latitude", iris.analysis.MAX
        )
        np.testing.assert_array_equal(
            cube.coord("surface_altitude").points,
            np.array([112, 113, 114, 115, 116, 117]),
        )

        np.testing.assert_array_equal(
            cube.coord("surface_altitude").bounds,
            np.array(
                [
                    [100, 124],
                    [101, 125],
                    [102, 126],
                    [103, 127],
                    [104, 128],
                    [105, 129],
                ]
            ),
        )

        # Check collapsing over the whole coord still works
        cube = self.cube_with_aux_coord.collapsed(
            "altitude", iris.analysis.MAX
        )

        np.testing.assert_array_equal(
            cube.coord("surface_altitude").points, np.array([114])
        )

        np.testing.assert_array_equal(
            cube.coord("surface_altitude").bounds, np.array([[100, 129]])
        )

        cube = self.cube_with_aux_coord.collapsed(
            "grid_longitude", iris.analysis.MAX
        )

        np.testing.assert_array_equal(
            cube.coord("surface_altitude").points,
            np.array([102, 108, 114, 120, 126]),
        )

        np.testing.assert_array_equal(
            cube.coord("surface_altitude").bounds,
            np.array(
                [[100, 105], [106, 111], [112, 117], [118, 123], [124, 129]]
            ),
        )


class TestAggregator_mdtol_keyword(tests.IrisTest):
    def setUp(self):
        data = ma.array(
            [[1, 2], [4, 5]],
            dtype=np.float32,
            mask=[[False, True], [False, True]],
        )
        cube = iris.cube.Cube(data, long_name="test_data", units="1")
        lat_coord = iris.coords.DimCoord(
            np.array([1, 2], dtype=np.float32), long_name="lat", units="1"
        )
        lon_coord = iris.coords.DimCoord(
            np.array([3, 4], dtype=np.float32), long_name="lon", units="1"
        )
        cube.add_dim_coord(lat_coord, 0)
        cube.add_dim_coord(lon_coord, 1)
        self.cube = cube

    def test_single_coord_no_mdtol(self):
        collapsed = self.cube.collapsed(
            self.cube.coord("lat"), iris.analysis.MEAN
        )
        t = ma.array([2.5, 5.0], mask=[False, True])
        self.assertMaskedArrayEqual(collapsed.data, t)

    def test_single_coord_mdtol(self):
        self.cube.data.mask = np.array([[False, True], [False, False]])
        collapsed = self.cube.collapsed(
            self.cube.coord("lat"), iris.analysis.MEAN, mdtol=0.5
        )
        t = ma.array([2.5, 5], mask=[False, False])
        self.assertMaskedArrayEqual(collapsed.data, t)

    def test_single_coord_mdtol_alt(self):
        self.cube.data.mask = np.array([[False, True], [False, False]])
        collapsed = self.cube.collapsed(
            self.cube.coord("lat"), iris.analysis.MEAN, mdtol=0.4
        )
        t = ma.array([2.5, 5], mask=[False, True])
        self.assertMaskedArrayEqual(collapsed.data, t)

    def test_multi_coord_no_mdtol(self):
        collapsed = self.cube.collapsed(
            [self.cube.coord("lat"), self.cube.coord("lon")],
            iris.analysis.MEAN,
        )
        t = np.array(2.5)
        self.assertArrayEqual(collapsed.data, t)

    def test_multi_coord_mdtol(self):
        collapsed = self.cube.collapsed(
            [self.cube.coord("lat"), self.cube.coord("lon")],
            iris.analysis.MEAN,
            mdtol=0.4,
        )
        t = ma.array(2.5, mask=True)
        self.assertMaskedArrayEqual(collapsed.data, t)


class TestAggregators(tests.IrisTest):
    def _check_collapsed_percentile(
        self,
        cube,
        percents,
        collapse_coord,
        expected_result,
        CML_filename=None,
        **kwargs,
    ):
        cube_data_type = type(cube.data)
        expected_result = np.array(expected_result, dtype=np.float32)
        result = cube.collapsed(
            collapse_coord,
            iris.analysis.PERCENTILE,
            percent=percents,
            **kwargs,
        )
        np.testing.assert_array_almost_equal(result.data, expected_result)
        self.assertEqual(type(result.data), cube_data_type)
        if CML_filename is not None:
            self.assertCML(result, ("analysis", CML_filename), checksum=False)

    def _check_percentile(
        self, data, axis, percents, expected_result, **kwargs
    ):
        result = iris.analysis._percentile(data, axis, percents, **kwargs)
        np.testing.assert_array_almost_equal(result, expected_result)
        self.assertEqual(type(result), type(expected_result))

    def test_percentile_1d_25_percent(self):
        cube = tests.stock.simple_1d()
        self._check_collapsed_percentile(
            cube, 25, "foo", 2.5, CML_filename="first_quartile_foo_1d.cml"
        )

    def test_percentile_1d_75_percent(self):
        cube = tests.stock.simple_1d()
        self._check_collapsed_percentile(
            cube, 75, "foo", 7.5, CML_filename="third_quartile_foo_1d.cml"
        )

    def test_fast_percentile_1d_25_percent(self):
        cube = tests.stock.simple_1d()
        self._check_collapsed_percentile(
            cube,
            25,
            "foo",
            2.5,
            fast_percentile_method=True,
            CML_filename="first_quartile_foo_1d_fast_percentile.cml",
        )

    def test_fast_percentile_1d_75_percent(self):
        cube = tests.stock.simple_1d()
        self._check_collapsed_percentile(
            cube,
            75,
            "foo",
            7.5,
            fast_percentile_method=True,
            CML_filename="third_quartile_foo_1d_fast_percentile.cml",
        )

    def test_fast_percentile_1d_75_percent_masked_type_no_mask(self):
        cube = tests.stock.simple_1d()
        cube.data = ma.MaskedArray(cube.data)
        self._check_collapsed_percentile(
            cube,
            75,
            "foo",
            7.5,
            fast_percentile_method=True,
            CML_filename="third_quartile_foo_1d_fast_percentile.cml",
        )

    def test_percentile_2d_single_coord(self):
        cube = tests.stock.simple_2d()
        self._check_collapsed_percentile(
            cube,
            25,
            "foo",
            [0.75, 4.75, 8.75],
            CML_filename="first_quartile_foo_2d.cml",
        )

    def test_percentile_2d_two_coords(self):
        cube = tests.stock.simple_2d()
        self._check_collapsed_percentile(
            cube,
            25,
            ["foo", "bar"],
            [2.75],
            CML_filename="first_quartile_foo_bar_2d.cml",
        )

    def test_fast_percentile_2d_single_coord(self):
        cube = tests.stock.simple_2d()
        self._check_collapsed_percentile(
            cube,
            25,
            "foo",
            [0.75, 4.75, 8.75],
            fast_percentile_method=True,
            CML_filename="first_quartile_foo_2d_fast_percentile.cml",
        )

    def test_fast_percentile_2d_two_coords(self):
        cube = tests.stock.simple_2d()
        self._check_collapsed_percentile(
            cube,
            25,
            ["foo", "bar"],
            [2.75],
            fast_percentile_method=True,
            CML_filename="first_quartile_foo_bar_2d_fast_percentile.cml",
        )

    def test_fast_percentile_2d_single_coord_masked_type_no_mask(self):
        cube = tests.stock.simple_2d()
        cube.data = ma.MaskedArray(cube.data)
        self._check_collapsed_percentile(
            cube,
            25,
            "foo",
            [0.75, 4.75, 8.75],
            fast_percentile_method=True,
            CML_filename="first_quartile_foo_2d_fast_percentile.cml",
        )

    def test_fast_percentile_2d_two_coords_masked_type_no_mask(self):
        cube = tests.stock.simple_2d()
        cube.data = ma.MaskedArray(cube.data)
        self._check_collapsed_percentile(
            cube,
            25,
            ["foo", "bar"],
            [2.75],
            fast_percentile_method=True,
            CML_filename="first_quartile_foo_bar_2d_fast_percentile.cml",
        )

    def test_percentile_3d(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [
                [6.0, 7.0, 8.0, 9.0],
                [10.0, 11.0, 12.0, 13.0],
                [14.0, 15.0, 16.0, 17.0],
            ],
            dtype=np.float32,
        )
        self._check_percentile(array_3d, 0, 50, expected_result)

    def test_fast_percentile_3d(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [
                [6.0, 7.0, 8.0, 9.0],
                [10.0, 11.0, 12.0, 13.0],
                [14.0, 15.0, 16.0, 17.0],
            ],
            dtype=np.float32,
        )
        self._check_percentile(
            array_3d, 0, 50, expected_result, fast_percentile_method=True
        )

    def test_percentile_3d_axis_one(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [[4.0, 5.0, 6.0, 7.0], [16.0, 17.0, 18.0, 19.0]], dtype=np.float32
        )

        self._check_percentile(array_3d, 1, 50, expected_result)

    def test_fast_percentile_3d_axis_one(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [[4.0, 5.0, 6.0, 7.0], [16.0, 17.0, 18.0, 19.0]], dtype=np.float32
        )

        self._check_percentile(
            array_3d, 1, 50, expected_result, fast_percentile_method=True
        )

    def test_fast_percentile_3d_axis_one_masked_type_no_mask(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        array_3d = np.ma.MaskedArray(array_3d)
        expected_result = ma.MaskedArray(
            [[4.0, 5.0, 6.0, 7.0], [16.0, 17.0, 18.0, 19.0]], dtype=np.float32
        )

        self._check_percentile(
            array_3d, 1, 50, expected_result, fast_percentile_method=True
        )

    def test_percentile_3d_axis_two(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [[1.5, 5.5, 9.5], [13.5, 17.5, 21.5]], dtype=np.float32
        )

        self._check_percentile(array_3d, 2, 50, expected_result)

    def test_fast_percentile_3d_axis_two(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        expected_result = np.array(
            [[1.5, 5.5, 9.5], [13.5, 17.5, 21.5]], dtype=np.float32
        )

        self._check_percentile(
            array_3d, 2, 50, expected_result, fast_percentile_method=True
        )

    def test_fast_percentile_3d_axis_two_masked_type_no_mask(self):
        array_3d = np.arange(24, dtype=np.int32).reshape((2, 3, 4))
        array_3d = ma.MaskedArray(array_3d)
        expected_result = ma.MaskedArray(
            [[1.5, 5.5, 9.5], [13.5, 17.5, 21.5]], dtype=np.float32
        )

        self._check_percentile(
            array_3d, 2, 50, expected_result, fast_percentile_method=True
        )

    def test_percentile_3d_masked(self):
        cube = tests.stock.simple_3d_mask()
        expected_result = [
            [12.0, 13.0, 14.0, 15.0],
            [16.0, 17.0, 18.0, 19.0],
            [20.0, 18.0, 19.0, 20.0],
        ]

        self._check_collapsed_percentile(
            cube,
            75,
            "wibble",
            expected_result,
            CML_filename="last_quartile_foo_3d_masked.cml",
        )

    def test_fast_percentile_3d_masked_type_masked(self):
        cube = tests.stock.simple_3d_mask()
        msg = "Cannot use fast np.percentile method with masked array."

        with self.assertRaisesRegex(TypeError, msg):
            cube.collapsed(
                "wibble",
                iris.analysis.PERCENTILE,
                percent=75,
                fast_percentile_method=True,
            )

    def test_percentile_3d_notmasked(self):
        cube = tests.stock.simple_3d()
        expected_result = [
            [9.0, 10.0, 11.0, 12.0],
            [13.0, 14.0, 15.0, 16.0],
            [17.0, 18.0, 19.0, 20.0],
        ]

        self._check_collapsed_percentile(
            cube,
            75,
            "wibble",
            expected_result,
            CML_filename="last_quartile_foo_3d_notmasked.cml",
        )

    def test_fast_percentile_3d_notmasked(self):
        cube = tests.stock.simple_3d()
        expected_result = [
            [9.0, 10.0, 11.0, 12.0],
            [13.0, 14.0, 15.0, 16.0],
            [17.0, 18.0, 19.0, 20.0],
        ]

        self._check_collapsed_percentile(
            cube,
            75,
            "wibble",
            expected_result,
            fast_percentile_method=True,
            CML_filename="last_quartile_foo_3d_notmasked_fast_percentile.cml",
        )

    def test_proportion(self):
        cube = tests.stock.simple_1d()
        assert np.any(cube.data >= 5)
        gt5 = cube.collapsed(
            "foo", iris.analysis.PROPORTION, function=lambda val: val >= 5
        )
        np.testing.assert_array_almost_equal(gt5.data, np.array([6 / 11.0]))
        self.assertCML(
            gt5, ("analysis", "proportion_foo_1d.cml"), checksum=False
        )

    def test_proportion_2d(self):
        cube = tests.stock.simple_2d()

        gt6 = cube.collapsed(
            "foo", iris.analysis.PROPORTION, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([0, 0.5, 1], dtype=np.float32)
        )
        self.assertCML(
            gt6, ("analysis", "proportion_foo_2d.cml"), checksum=False
        )

        gt6 = cube.collapsed(
            "bar", iris.analysis.PROPORTION, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([1 / 3, 1 / 3, 2 / 3, 2 / 3], dtype=np.float32)
        )
        self.assertCML(
            gt6, ("analysis", "proportion_bar_2d.cml"), checksum=False
        )

        gt6 = cube.collapsed(
            ("foo", "bar"),
            iris.analysis.PROPORTION,
            function=lambda val: val >= 6,
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([0.5], dtype=np.float32)
        )
        self.assertCML(
            gt6, ("analysis", "proportion_foo_bar_2d.cml"), checksum=False
        )

        # mask the data
        cube.data = ma.array(cube.data, mask=cube.data % 2)
        cube.data.mask[1, 2] = True
        gt6_masked = cube.collapsed(
            "bar", iris.analysis.PROPORTION, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6_masked.data,
            ma.array(
                [1 / 3, None, 1 / 2, None],
                mask=[False, True, False, True],
                dtype=np.float32,
            ),
        )
        self.assertCML(
            gt6_masked,
            ("analysis", "proportion_foo_2d_masked.cml"),
            checksum=False,
        )

    def test_count(self):
        cube = tests.stock.simple_1d()
        gt5 = cube.collapsed(
            "foo", iris.analysis.COUNT, function=lambda val: val >= 5
        )
        np.testing.assert_array_almost_equal(gt5.data, np.array([6]))
        gt5.data = gt5.data.astype("i8")
        self.assertCML(gt5, ("analysis", "count_foo_1d.cml"), checksum=False)

    def test_count_2d(self):
        cube = tests.stock.simple_2d()

        gt6 = cube.collapsed(
            "foo", iris.analysis.COUNT, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([0, 2, 4], dtype=np.float32)
        )
        gt6.data = gt6.data.astype("i8")
        self.assertCML(gt6, ("analysis", "count_foo_2d.cml"), checksum=False)

        gt6 = cube.collapsed(
            "bar", iris.analysis.COUNT, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([1, 1, 2, 2], dtype=np.float32)
        )
        gt6.data = gt6.data.astype("i8")
        self.assertCML(gt6, ("analysis", "count_bar_2d.cml"), checksum=False)

        gt6 = cube.collapsed(
            ("foo", "bar"), iris.analysis.COUNT, function=lambda val: val >= 6
        )
        np.testing.assert_array_almost_equal(
            gt6.data, np.array([6], dtype=np.float32)
        )
        gt6.data = gt6.data.astype("i8")
        self.assertCML(
            gt6, ("analysis", "count_foo_bar_2d.cml"), checksum=False
        )

    def test_weighted_sum_consistency(self):
        # weighted sum with unit weights should be the same as a sum
        cube = tests.stock.simple_1d()
        normal_sum = cube.collapsed("foo", iris.analysis.SUM)
        weights = np.ones_like(cube.data)
        weighted_sum = cube.collapsed(
            "foo", iris.analysis.SUM, weights=weights
        )
        self.assertArrayAlmostEqual(normal_sum.data, weighted_sum.data)

    def test_weighted_sum_1d(self):
        # verify 1d weighted sum is correct
        cube = tests.stock.simple_1d()
        weights = np.array(
            [0.05, 0.05, 0.1, 0.1, 0.2, 0.3, 0.2, 0.1, 0.1, 0.05, 0.05]
        )
        result = cube.collapsed("foo", iris.analysis.SUM, weights=weights)
        self.assertAlmostEqual(result.data, 6.5)
        self.assertCML(
            result, ("analysis", "sum_weighted_1d.cml"), checksum=False
        )

    def test_weighted_sum_2d(self):
        # verify 2d weighted sum is correct
        cube = tests.stock.simple_2d()
        weights = np.array([0.3, 0.4, 0.3])
        weights = iris.util.broadcast_to_shape(weights, cube.shape, [0])
        result = cube.collapsed("bar", iris.analysis.SUM, weights=weights)
        self.assertArrayAlmostEqual(
            result.data, np.array([4.0, 5.0, 6.0, 7.0])
        )
        self.assertCML(
            result, ("analysis", "sum_weighted_2d.cml"), checksum=False
        )

    def test_weighted_rms(self):
        cube = tests.stock.simple_2d()
        # modify cube data so that the results are nice numbers
        cube.data = np.array(
            [[4, 7, 10, 8], [21, 30, 12, 24], [14, 16, 20, 8]],
            dtype=np.float64,
        )
        weights = np.array(
            [[1, 4, 3, 2], [6, 4.5, 1.5, 3], [2, 1, 1.5, 0.5]],
            dtype=np.float64,
        )
        expected_result = np.array([8.0, 24.0, 16.0])
        result = cube.collapsed("foo", iris.analysis.RMS, weights=weights)
        self.assertArrayAlmostEqual(result.data, expected_result)
        self.assertCML(
            result, ("analysis", "rms_weighted_2d.cml"), checksum=False
        )


@tests.skip_data
class TestRotatedPole(tests.IrisTest):
    def _check_both_conversions(self, cube, index):
        rlons, rlats = iris.analysis.cartography.get_xy_grids(cube)
        rcs = cube.coord_system("RotatedGeogCS")
        x, y = iris.analysis.cartography.unrotate_pole(
            rlons,
            rlats,
            rcs.grid_north_pole_longitude,
            rcs.grid_north_pole_latitude,
        )
        self.assertDataAlmostEqual(
            x, ("analysis", "rotated_pole.{}.x.json".format(index))
        )
        self.assertDataAlmostEqual(
            y, ("analysis", "rotated_pole.{}.y.json".format(index))
        )
        self.assertDataAlmostEqual(
            rlons, ("analysis", "rotated_pole.{}.rlon.json".format(index))
        )
        self.assertDataAlmostEqual(
            rlats, ("analysis", "rotated_pole.{}.rlat.json".format(index))
        )

    def test_all(self):
        path = tests.get_data_path(("PP", "ukVorog", "ukv_orog_refonly.pp"))
        master_cube = iris.load_cube(path)

        # Check overall behaviour.
        cube = master_cube[::10, ::10]
        self._check_both_conversions(cube, 0)

        # Check numerical stability.
        cube = master_cube[210:238, 424:450]
        self._check_both_conversions(cube, 1)

    def test_unrotate_nd(self):
        rlons = np.array([[350.0, 352.0], [350.0, 352.0]])
        rlats = np.array([[-5.0, -0.0], [-4.0, -1.0]])

        resx, resy = iris.analysis.cartography.unrotate_pole(
            rlons, rlats, 178.0, 38.0
        )

        # Solutions derived by proj4 direct.
        solx = np.array(
            [[-16.42176094, -14.85892262], [-16.71055023, -14.58434624]]
        )
        soly = np.array(
            [[46.00724251, 51.29188893], [46.98728486, 50.30706042]]
        )

        self.assertArrayAlmostEqual(resx, solx)
        self.assertArrayAlmostEqual(resy, soly)

    def test_unrotate_1d(self):
        rlons = np.array([350.0, 352.0, 354.0, 356.0])
        rlats = np.array([-5.0, -0.0, 5.0, 10.0])

        resx, resy = iris.analysis.cartography.unrotate_pole(
            rlons.flatten(), rlats.flatten(), 178.0, 38.0
        )

        # Solutions derived by proj4 direct.
        solx = np.array(
            [-16.42176094, -14.85892262, -12.88946157, -10.35078336]
        )
        soly = np.array([46.00724251, 51.29188893, 56.55031485, 61.77015703])

        self.assertArrayAlmostEqual(resx, solx)
        self.assertArrayAlmostEqual(resy, soly)

    def test_rotate_nd(self):
        rlons = np.array([[350.0, 351.0], [352.0, 353.0]])
        rlats = np.array([[10.0, 15.0], [20.0, 25.0]])

        resx, resy = iris.analysis.cartography.rotate_pole(
            rlons, rlats, 20.0, 80.0
        )

        # Solutions derived by proj4 direct.
        solx = np.array(
            [[148.69672569, 149.24727087], [149.79067025, 150.31754368]]
        )
        soly = np.array(
            [[18.60905789, 23.67749384], [28.74419024, 33.8087963]]
        )

        self.assertArrayAlmostEqual(resx, solx)
        self.assertArrayAlmostEqual(resy, soly)

    def test_rotate_1d(self):
        rlons = np.array([350.0, 351.0, 352.0, 353.0])
        rlats = np.array([10.0, 15.0, 20.0, 25.0])

        resx, resy = iris.analysis.cartography.rotate_pole(
            rlons.flatten(), rlats.flatten(), 20.0, 80.0
        )

        # Solutions derived by proj4 direct.
        solx = np.array(
            [148.69672569, 149.24727087, 149.79067025, 150.31754368]
        )
        soly = np.array([18.60905789, 23.67749384, 28.74419024, 33.8087963])

        self.assertArrayAlmostEqual(resx, solx)
        self.assertArrayAlmostEqual(resy, soly)


@tests.skip_data
class TestAreaWeights(tests.IrisTest):
    def test_area_weights(self):
        small_cube = iris.tests.stock.simple_pp()
        # Get offset, subsampled region: small enough to test against literals
        small_cube = small_cube[10:, 35:]
        small_cube = small_cube[::8, ::8]
        small_cube = small_cube[:5, :4]
        # pre-check non-data properties
        self.assertCML(
            small_cube,
            ("analysis", "areaweights_original.cml"),
            checksum=False,
        )

        # check area-weights values
        small_cube.coord("latitude").guess_bounds()
        small_cube.coord("longitude").guess_bounds()
        area_weights = iris.analysis.cartography.area_weights(small_cube)
        expected_results = np.array(
            [
                [3.11955866e12, 3.11956008e12, 3.11955866e12, 3.11956008e12],
                [5.21951065e12, 5.21951303e12, 5.21951065e12, 5.21951303e12],
                [6.68991281e12, 6.68991585e12, 6.68991281e12, 6.68991585e12],
                [7.35341305e12, 7.35341640e12, 7.35341305e12, 7.35341640e12],
                [7.12998335e12, 7.12998660e12, 7.12998335e12, 7.12998660e12],
            ],
            dtype=np.float64,
        )
        self.assertArrayAllClose(area_weights, expected_results, rtol=1e-8)

        # Check there was no residual change
        small_cube.coord("latitude").bounds = None
        small_cube.coord("longitude").bounds = None
        self.assertCML(
            small_cube,
            ("analysis", "areaweights_original.cml"),
            checksum=False,
        )


@tests.skip_data
class TestAreaWeightGeneration(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.realistic_4d()

    def test_area_weights_std(self):
        # weights for stock 4d data
        weights = iris.analysis.cartography.area_weights(self.cube)
        self.assertEqual(weights.shape, self.cube.shape)

    def test_area_weights_order(self):
        # weights for data with dimensions in a different order
        order = [3, 2, 1, 0]  # (lon, lat, level, time)
        self.cube.transpose(order)
        weights = iris.analysis.cartography.area_weights(self.cube)
        self.assertEqual(weights.shape, self.cube.shape)

    def test_area_weights_non_adjacent(self):
        # weights for cube with non-adjacent latitude/longitude dimensions
        order = [0, 3, 1, 2]  # (time, lon, level, lat)
        self.cube.transpose(order)
        weights = iris.analysis.cartography.area_weights(self.cube)
        self.assertEqual(weights.shape, self.cube.shape)

    def test_area_weights_scalar_latitude(self):
        # weights for cube with a scalar latitude dimension
        cube = self.cube[:, :, 0, :]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_scalar_longitude(self):
        # weights for cube with a scalar longitude dimension
        cube = self.cube[:, :, :, 0]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_scalar(self):
        # weights for cube with scalar latitude and longitude dimensions
        cube = self.cube[:, :, 0, 0]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_singleton_latitude(self):
        # singleton (1-point) latitude dimension
        cube = self.cube[:, :, 0:1, :]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_singleton_longitude(self):
        # singleton (1-point) longitude dimension
        cube = self.cube[:, :, :, 0:1]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_singletons(self):
        # singleton (1-point) latitude and longitude dimensions
        cube = self.cube[:, :, 0:1, 0:1]
        weights = iris.analysis.cartography.area_weights(cube)
        self.assertEqual(weights.shape, cube.shape)

    def test_area_weights_normalized(self):
        # normalized area weights must sum to one over lat/lon dimensions.
        weights = iris.analysis.cartography.area_weights(
            self.cube, normalize=True
        )
        sumweights = weights.sum(axis=3).sum(axis=2)  # sum over lon and lat
        self.assertArrayAlmostEqual(sumweights, 1)

    def test_area_weights_non_contiguous(self):
        # Slice the cube so that we have non-contiguous longitude
        # bounds.
        ind = (0, 1, 2, -3, -2, -1)
        cube = self.cube[..., ind]
        weights = iris.analysis.cartography.area_weights(cube)
        expected = iris.analysis.cartography.area_weights(self.cube)[..., ind]
        self.assertArrayEqual(weights, expected)

    def test_area_weights_no_lon_bounds(self):
        self.cube.coord("grid_longitude").bounds = None
        with self.assertRaises(ValueError):
            iris.analysis.cartography.area_weights(self.cube)

    def test_area_weights_no_lat_bounds(self):
        self.cube.coord("grid_latitude").bounds = None
        with self.assertRaises(ValueError):
            iris.analysis.cartography.area_weights(self.cube)


@tests.skip_data
class TestLatitudeWeightGeneration(tests.IrisTest):
    def setUp(self):
        path = iris.tests.get_data_path(
            ["NetCDF", "rotated", "xyt", "small_rotPole_precipitation.nc"]
        )
        self.cube = iris.load_cube(path)
        self.cube_dim_lat = self.cube.copy()
        self.cube_dim_lat.remove_coord("latitude")
        self.cube_dim_lat.remove_coord("longitude")
        # The 2d cubes are unrealistic, you would not want to weight by
        # anything other than grid latitude in real-world scenarios. However,
        # the technical details are suitable for testing purposes, providing
        # a nice analog for a 2d latitude coordinate from a curvilinear grid.
        self.cube_aux_lat = self.cube.copy()
        self.cube_aux_lat.remove_coord("grid_latitude")
        self.cube_aux_lat.remove_coord("grid_longitude")
        self.lat1d = self.cube.coord("grid_latitude").points
        self.lat2d = self.cube.coord("latitude").points

    def test_cosine_latitude_weights_range(self):
        # check the range of returned values, needs a cube that spans the full
        # latitude range
        lat_coord = iris.coords.DimCoord(
            np.linspace(-90, 90, 73),
            standard_name="latitude",
            units=cf_units.Unit("degrees_north"),
        )
        cube = iris.cube.Cube(
            np.ones([73], dtype=np.float64), long_name="test_cube", units="1"
        )
        cube.add_dim_coord(lat_coord, 0)
        weights = iris.analysis.cartography.cosine_latitude_weights(cube)
        self.assertTrue(weights.max() <= 1)
        self.assertTrue(weights.min() >= 0)

    def test_cosine_latitude_weights_0d(self):
        # 0d latitude dimension (scalar coordinate)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_dim_lat[:, 0, :]
        )
        self.assertEqual(weights.shape, self.cube_dim_lat[:, 0, :].shape)
        self.assertAlmostEqual(
            weights[0, 0], np.cos(np.deg2rad(self.lat1d[0]))
        )

    def test_cosine_latitude_weights_1d_singleton(self):
        # singleton (1-point) 1d latitude coordinate (time, lat, lon)
        cube = self.cube_dim_lat[:, 0:1, :]
        weights = iris.analysis.cartography.cosine_latitude_weights(cube)
        self.assertEqual(weights.shape, cube.shape)
        self.assertAlmostEqual(
            weights[0, 0, 0], np.cos(np.deg2rad(self.lat1d[0]))
        )

    def test_cosine_latitude_weights_1d(self):
        # 1d latitude coordinate (time, lat, lon)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_dim_lat
        )
        self.assertEqual(weights.shape, self.cube.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, 0], np.cos(np.deg2rad(self.lat1d))
        )

    def test_cosine_latitude_weights_1d_latitude_first(self):
        # 1d latitude coordinate with latitude first (lat, time, lon)
        order = [1, 0, 2]  # (lat, time, lon)
        self.cube_dim_lat.transpose(order)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_dim_lat
        )
        self.assertEqual(weights.shape, self.cube_dim_lat.shape)
        self.assertArrayAlmostEqual(
            weights[:, 0, 0], np.cos(np.deg2rad(self.lat1d))
        )

    def test_cosine_latitude_weights_1d_latitude_last(self):
        # 1d latitude coordinate with latitude last (time, lon, lat)
        order = [0, 2, 1]  # (time, lon, lat)
        self.cube_dim_lat.transpose(order)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_dim_lat
        )
        self.assertEqual(weights.shape, self.cube_dim_lat.shape)
        self.assertArrayAlmostEqual(
            weights[0, 0, :], np.cos(np.deg2rad(self.lat1d))
        )

    def test_cosine_latitude_weights_2d_singleton1(self):
        # 2d latitude coordinate with first dimension singleton
        cube = self.cube_aux_lat[:, 0:1, :]
        weights = iris.analysis.cartography.cosine_latitude_weights(cube)
        self.assertEqual(weights.shape, cube.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, :], np.cos(np.deg2rad(self.lat2d[0:1, :]))
        )

    def test_cosine_latitude_weights_2d_singleton2(self):
        # 2d latitude coordinate with second dimension singleton
        cube = self.cube_aux_lat[:, :, 0:1]
        weights = iris.analysis.cartography.cosine_latitude_weights(cube)
        self.assertEqual(weights.shape, cube.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, :], np.cos(np.deg2rad(self.lat2d[:, 0:1]))
        )

    def test_cosine_latitude_weights_2d_singleton3(self):
        # 2d latitude coordinate with both dimensions singleton
        cube = self.cube_aux_lat[:, 0:1, 0:1]
        weights = iris.analysis.cartography.cosine_latitude_weights(cube)
        self.assertEqual(weights.shape, cube.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, :], np.cos(np.deg2rad(self.lat2d[0:1, 0:1]))
        )

    def test_cosine_latitude_weights_2d(self):
        # 2d latitude coordinate (time, lat, lon)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_aux_lat
        )
        self.assertEqual(weights.shape, self.cube_aux_lat.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, :], np.cos(np.deg2rad(self.lat2d))
        )

    def test_cosine_latitude_weights_2d_latitude_first(self):
        # 2d latitude coordinate with latitude first (lat, time, lon)
        order = [1, 0, 2]  # (lat, time, lon)
        self.cube_aux_lat.transpose(order)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_aux_lat
        )
        self.assertEqual(weights.shape, self.cube_aux_lat.shape)
        self.assertArrayAlmostEqual(
            weights[:, 0, :], np.cos(np.deg2rad(self.lat2d))
        )

    def test_cosine_latitude_weights_2d_latitude_last(self):
        # 2d latitude coordinate with latitude last (time, lon, lat)
        order = [0, 2, 1]  # (time, lon, lat)
        self.cube_aux_lat.transpose(order)
        weights = iris.analysis.cartography.cosine_latitude_weights(
            self.cube_aux_lat
        )
        self.assertEqual(weights.shape, self.cube_aux_lat.shape)
        self.assertArrayAlmostEqual(
            weights[0, :, :], np.cos(np.deg2rad(self.lat2d.T))
        )

    def test_cosine_latitude_weights_no_latitude(self):
        # no coordinate identified as latitude
        self.cube_dim_lat.remove_coord("grid_latitude")
        with self.assertRaises(ValueError):
            _ = iris.analysis.cartography.cosine_latitude_weights(
                self.cube_dim_lat
            )

    def test_cosine_latitude_weights_multiple_latitude(self):
        # two coordinates identified as latitude
        with self.assertRaises(ValueError):
            _ = iris.analysis.cartography.cosine_latitude_weights(self.cube)


class TestRollingWindow(tests.IrisTest):
    def setUp(self):
        # XXX Comes from test_aggregated_by
        cube = iris.cube.Cube(
            np.array([[6, 10, 12, 18], [8, 12, 14, 20], [18, 12, 10, 6]]),
            long_name="temperature",
            units="kelvin",
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.array([0, 5, 10], dtype=np.float64),
                "latitude",
                units="degrees",
            ),
            0,
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.array([0, 2, 4, 6], dtype=np.float64),
                "longitude",
                units="degrees",
            ),
            1,
        )

        self.cube = cube

    def test_non_mean_operator(self):
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MAX, window=2
        )
        expected_result = np.array(
            [[10, 12, 18], [12, 14, 20], [18, 12, 10]], dtype=np.float64
        )
        self.assertArrayEqual(expected_result, res_cube.data)

    def test_longitude_simple(self):
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=2
        )

        expected_result = np.array(
            [[8.0, 11.0, 15.0], [10.0, 13.0, 17.0], [15.0, 11.0, 8.0]],
            dtype=np.float64,
        )

        self.assertArrayEqual(expected_result, res_cube.data)

        self.assertCML(
            res_cube, ("analysis", "rolling_window", "simple_longitude.cml")
        )

        self.assertRaises(
            ValueError,
            self.cube.rolling_window,
            "longitude",
            iris.analysis.MEAN,
            window=0,
        )

    def test_longitude_masked(self):
        self.cube.data = ma.array(
            self.cube.data,
            mask=[
                [True, True, True, True],
                [True, False, True, True],
                [False, False, False, False],
            ],
        )
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=2
        )

        expected_result = np.ma.array(
            [[-99.0, -99.0, -99.0], [12.0, 12.0, -99.0], [15.0, 11.0, 8.0]],
            mask=[
                [True, True, True],
                [False, False, True],
                [False, False, False],
            ],
            dtype=np.float64,
        )

        self.assertMaskedArrayEqual(expected_result, res_cube.data)

    def test_longitude_circular(self):
        cube = self.cube
        cube.coord("longitude").circular = True
        self.assertRaises(
            iris.exceptions.NotYetImplementedError,
            self.cube.rolling_window,
            "longitude",
            iris.analysis.MEAN,
            window=0,
        )

    def test_different_length_windows(self):
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=4
        )

        expected_result = np.array([[11.5], [13.5], [11.5]], dtype=np.float64)

        self.assertArrayEqual(expected_result, res_cube.data)

        self.assertCML(
            res_cube, ("analysis", "rolling_window", "size_4_longitude.cml")
        )

        # Window too long:
        self.assertRaises(
            ValueError,
            self.cube.rolling_window,
            "longitude",
            iris.analysis.MEAN,
            window=6,
        )
        # Window too small:
        self.assertRaises(
            ValueError,
            self.cube.rolling_window,
            "longitude",
            iris.analysis.MEAN,
            window=0,
        )

    def test_bad_coordinate(self):
        self.assertRaises(
            KeyError,
            self.cube.rolling_window,
            "wibble",
            iris.analysis.MEAN,
            window=0,
        )

    def test_latitude_simple(self):
        res_cube = self.cube.rolling_window(
            "latitude", iris.analysis.MEAN, window=2
        )

        expected_result = np.array(
            [[7.0, 11.0, 13.0, 19.0], [13.0, 12.0, 12.0, 13.0]],
            dtype=np.float64,
        )

        self.assertArrayEqual(expected_result, res_cube.data)

        self.assertCML(
            res_cube, ("analysis", "rolling_window", "simple_latitude.cml")
        )

    def test_mean_with_weights_consistency(self):
        # equal weights should be the same as the mean with no weights
        wts = np.array([0.5, 0.5], dtype=np.float64)
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=2, weights=wts
        )
        expected_result = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=2
        )
        self.assertArrayEqual(expected_result.data, res_cube.data)

    def test_mean_with_weights(self):
        # rolling window mean with weights
        wts = np.array([0.1, 0.6, 0.3], dtype=np.float64)
        res_cube = self.cube.rolling_window(
            "longitude", iris.analysis.MEAN, window=3, weights=wts
        )
        expected_result = np.array(
            [[10.2, 13.6], [12.2, 15.6], [12.0, 9.0]], dtype=np.float64
        )
        # use almost equal to compare floats
        self.assertArrayAlmostEqual(expected_result, res_cube.data)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Code for fast loading of structured UM data.

This module defines which pp-field elements take part in structured loading,
and provides creation of :class:`BasicFieldCollation` objects from lists of
:class:`iris.fileformats.pp.PPField`.

"""

import itertools

import cftime
import numpy as np

from iris._lazy_data import as_lazy_data, multidim_lazy_stack
from iris.fileformats.um._optimal_array_structuring import (
    optimal_array_structure,
)


class BasicFieldCollation:
    """
    An object representing a group of UM fields with array structure that can
    be vectorized into a single cube.

    For example:

    Suppose we have a set of 28 fields repeating over 7 vertical levels for
    each of 4 different data times.  If a BasicFieldCollation is created to
    contain these, it can identify that this is a 4*7 regular array structure.

    This BasicFieldCollation will then have the following properties:

    * within 'element_arrays_and_dims' :
        Element 'blev' have the array shape (7,) and dims of (1,).
        Elements 't1' and 't2' have shape (4,) and dims (0,).
        The other elements (lbft, lbrsvd4 and lbuser5) all have scalar array
        values and dims=None.

    .. note::

        If no array structure is found, the element values are all
        either scalar or full-length 1-D vectors.

    """

    def __init__(self, fields):
        """
        Args:

        * fields (iterable of :class:`iris.fileformats.pp.PPField`):
            The fields in the collation.

        """
        self._fields = tuple(fields)
        self._data_cache = None
        assert len(self.fields) > 0
        self._structure_calculated = False
        self._vector_dims_shape = None
        self._primary_dimension_elements = None
        self._element_arrays_and_dims = None

    @property
    def fields(self):
        return self._fields

    @property
    def data(self):
        if not self._structure_calculated:
            self._calculate_structure()
        if self._data_cache is None:
            stack = np.empty(self.vector_dims_shape, "object")
            for nd_index, field in zip(
                np.ndindex(self.vector_dims_shape), self.fields
            ):
                stack[nd_index] = as_lazy_data(field._data)
            self._data_cache = multidim_lazy_stack(stack)
        return self._data_cache

    def core_data(self):
        return self.data

    @property
    def realised_dtype(self):
        return np.result_type(
            *[field.realised_dtype for field in self._fields]
        )

    @property
    def data_proxy(self):
        return self.data

    @property
    def bmdi(self):
        bmdis = set([f.bmdi for f in self.fields])
        if len(bmdis) != 1:
            raise ValueError("Multiple bmdi values defined in FieldCollection")
        return bmdis.pop()

    @property
    def vector_dims_shape(self):
        """The shape of the array structure."""
        if not self._structure_calculated:
            self._calculate_structure()
        return self._vector_dims_shape

    @property
    def _UNUSED_primary_dimension_elements(self):
        """A set of names of the elements which are array dimensions."""
        if not self._structure_calculated:
            self._calculate_structure()
        return self._primary_dimension_elements

    @property
    def element_arrays_and_dims(self):
        """
        Value arrays for vector metadata elements.

        A dictionary mapping element_name: (value_array, dims).

        The arrays are reduced to their minimum dimensions.  A scalar array
        has an associated 'dims' of None (instead of an empty tuple).

        """
        if not self._structure_calculated:
            self._calculate_structure()
        return self._element_arrays_and_dims

    def _field_vector_element_arrays(self):
        """Define the field components used in the structure analysis."""
        # Define functions to make t1 and t2 values as date-time tuples.
        # These depend on header version (PPField2 has no seconds values).
        def t1_fn(fld):
            return (
                fld.lbyr,
                fld.lbmon,
                fld.lbdat,
                fld.lbhr,
                fld.lbmin,
                getattr(fld, "lbsec", 0),
            )

        def t2_fn(fld):
            return (
                fld.lbyrd,
                fld.lbmond,
                fld.lbdatd,
                fld.lbhrd,
                fld.lbmind,
                getattr(fld, "lbsecd", 0),
            )

        # Return a list of (name, array) for the vectorizable elements.
        component_arrays = [
            ("t1", np.array([t1_fn(fld) for fld in self.fields])),
            ("t2", np.array([t2_fn(fld) for fld in self.fields])),
            ("lbft", np.array([fld.lbft for fld in self.fields])),
            ("blev", np.array([fld.blev for fld in self.fields])),
            ("lblev", np.array([fld.lblev for fld in self.fields])),
            ("bhlev", np.array([fld.bhlev for fld in self.fields])),
            ("bhrlev", np.array([fld.bhrlev for fld in self.fields])),
            ("brsvd1", np.array([fld.brsvd[0] for fld in self.fields])),
            ("brsvd2", np.array([fld.brsvd[1] for fld in self.fields])),
            ("brlev", np.array([fld.brlev for fld in self.fields])),
        ]
        return component_arrays

    # Static factors for the _time_comparable_int routine (seconds per period).
    _TIME_ELEMENT_MULTIPLIERS = np.cumprod([1, 60, 60, 24, 31, 12])[::-1]

    def _time_comparable_int(self, yr, mon, dat, hr, min, sec):
        """
        Return a single unique number representing a date-time tuple.

        This calculation takes no account of the time field's real calendar,
        instead giving every month 31 days, which preserves the required
        time ordering.

        """
        elements = np.array((yr, mon, dat, hr, min, sec))
        return np.sum(elements * self._TIME_ELEMENT_MULTIPLIERS)

    def _calculate_structure(self):
        # Make value arrays for the vectorisable field elements.
        element_definitions = self._field_vector_element_arrays()

        # Identify the vertical elements and payload.
        blev_array = dict(element_definitions).get("blev")
        vertical_elements = (
            "lblev",
            "bhlev",
            "bhrlev",
            "brsvd1",
            "brsvd2",
            "brlev",
        )

        # Make an ordering copy.
        ordering_definitions = element_definitions[:]
        # Replace time value tuples with integers and bind the vertical
        # elements to the (expected) primary vertical element "blev".
        for index, (name, array) in enumerate(ordering_definitions):
            if name in ("t1", "t2"):
                array = np.array(
                    [self._time_comparable_int(*tuple(val)) for val in array]
                )
                ordering_definitions[index] = (name, array)
            if name in vertical_elements and blev_array is not None:
                ordering_definitions[index] = (name, blev_array)

        # Perform the main analysis: get vector dimensions, elements, arrays.
        (
            dims_shape,
            primary_elements,
            vector_element_arrays_and_dims,
        ) = optimal_array_structure(ordering_definitions, element_definitions)

        # Replace time tuples in the result with real datetime-like values.
        # N.B. so we *don't* do this on the whole (expanded) input arrays.
        for name in ("t1", "t2"):
            if name in vector_element_arrays_and_dims:
                arr, dims = vector_element_arrays_and_dims[name]
                arr_shape = arr.shape[:-1]
                extra_length = arr.shape[-1]
                # Flatten out the array apart from the last dimension,
                # convert to cftime objects, then reshape back.
                arr = np.array(
                    [
                        cftime.datetime(*args)
                        for args in arr.reshape(-1, extra_length)
                    ]
                ).reshape(arr_shape)
                vector_element_arrays_and_dims[name] = (arr, dims)

        # Write the private cache values, exposed as public properties.
        self._vector_dims_shape = dims_shape
        self._primary_dimension_elements = primary_elements
        self._element_arrays_and_dims = vector_element_arrays_and_dims
        # Do all this only once.
        self._structure_calculated = True


def _um_collation_key_function(field):
    """
    Standard collation key definition for fast structured field loading.

    The elements used here are the minimum sufficient to define the
    'phenomenon', as described for :meth:`group_structured_fields`.

    """
    return (
        field.lbuser[3],  # stashcode first
        field.lbproc,  # then stats processing
        field.lbuser[6],  # then model
        field.lbuser[4],  # then pseudo-level : this one is a KLUDGE.
    )
    # NOTE: including pseudo-level here makes it treat different pseudo-levels
    # as different phenomena.  These will later be merged in the "ordinary"
    # post-load merge.
    # The current structured-load code fails to handle multiple pseudo-levels
    # correctly: because pseudo-level is not on in its list of "things that may
    # vary within a phenomenon", it will create a scalar pseudo-level
    # coordinate when it should have been a vector of values.
    # This kludge fixes that error, but it is inefficient because it bypasses
    # the structured load, producing n-levels times more 'raw' cubes.
    # It will also fail if any phenomenon occurs with multiple 'normal' levels
    # (i.e. lblev) *and* pseudo-levels (lbuser[4]).
    # TODO: it should be fairly easy to do this properly -- i.e. create a
    # vector pseudo-level coordinate directly in the structured load analysis.


def group_structured_fields(
    field_iterator, collation_class=BasicFieldCollation, **collation_kwargs
):
    """
    Collect structured fields into identified groups whose fields can be
    combined to form a single cube.

    Args:

    * field_iterator (iterator of :class:`iris.fileformats.pp.PPField`):
        A source of PP or FF fields.  N.B. order is significant.

    Kwargs:

    * collation_class (class):
        Type of collation wrapper to create from each group of fields.
    * collation_kwargs (dict):
        Additional constructor keywords for collation creation.

    The function sorts and collates on phenomenon-relevant metadata only,
    defined as the field components: 'lbuser[3]' (stash), 'lbproc' (statistic),
    'lbuser[6]' (model).
    Each distinct combination of these defines a specific phenomenon (or
    statistical aggregation of one), and those fields appear as a single
    iteration result.

    Implicitly, within each result group, *all* other metadata components
    should be either:

    *  the same for all fields,
    *  completely irrelevant, or
    *  used by a vectorised rule function (such as
       :func:`iris.fileformats.pp_load_rules._convert_time_coords`).

    Returns:
        A generator of 'collation_class' objects, each of which contains a
        single collated group from the input fields.

    .. note::

         At present, fields with different values of 'lbuser[4]' (pseudo-level)
         are *also* treated as different phenomena.  This is a temporary fix,
         standing in place for a more correct handling of pseudo-levels.

    """
    _fields = sorted(field_iterator, key=_um_collation_key_function)
    for _, fields in itertools.groupby(_fields, _um_collation_key_function):
        yield collation_class(tuple(fields), **collation_kwargs)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coord_systems.VerticalPerspective` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import cartopy.crs as ccrs
from iris.coord_systems import GeogCS, VerticalPerspective


class Test(tests.IrisTest):
    def setUp(self):
        self.latitude_of_projection_origin = 0.0
        self.longitude_of_projection_origin = 0.0
        self.perspective_point_height = 38204820000.0
        self.false_easting = 0.0
        self.false_northing = 0.0

        self.semi_major_axis = 6377563.396
        self.semi_minor_axis = 6356256.909
        self.ellipsoid = GeogCS(self.semi_major_axis, self.semi_minor_axis)
        self.globe = ccrs.Globe(
            semimajor_axis=self.semi_major_axis,
            semiminor_axis=self.semi_minor_axis,
            ellipse=None,
        )

        # Actual and expected coord system can be re-used for
        # VerticalPerspective.test_crs_creation and test_projection_creation.
        self.expected = ccrs.NearsidePerspective(
            central_longitude=self.longitude_of_projection_origin,
            central_latitude=self.latitude_of_projection_origin,
            satellite_height=self.perspective_point_height,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            globe=self.globe,
        )
        self.vp_cs = VerticalPerspective(
            self.latitude_of_projection_origin,
            self.longitude_of_projection_origin,
            self.perspective_point_height,
            self.false_easting,
            self.false_northing,
            self.ellipsoid,
        )

    def test_crs_creation(self):
        res = self.vp_cs.as_cartopy_crs()
        self.assertEqual(res, self.expected)

    def test_projection_creation(self):
        res = self.vp_cs.as_cartopy_projection()
        self.assertEqual(res, self.expected)

    def test_set_optional_args(self):
        # Check that setting the optional (non-ellipse) args works.
        crs = VerticalPerspective(
            0, 0, 1000, false_easting=100, false_northing=-203.7
        )
        self.assertEqualAndKind(crs.false_easting, 100.0)
        self.assertEqualAndKind(crs.false_northing, -203.7)

    def _check_crs_defaults(self, crs):
        # Check for property defaults when no kwargs options were set.
        # NOTE: except ellipsoid, which is done elsewhere.
        self.assertEqualAndKind(crs.false_easting, 0.0)
        self.assertEqualAndKind(crs.false_northing, 0.0)

    def test_no_optional_args(self):
        # Check expected defaults with no optional args.
        crs = VerticalPerspective(0, 0, 1000)
        self._check_crs_defaults(crs)

    def test_optional_args_None(self):
        # Check expected defaults with optional args=None.
        crs = VerticalPerspective(
            0, 0, 1000, false_easting=None, false_northing=None
        )
        self._check_crs_defaults(crs)


if __name__ == "__main__":
    tests.main()



"""
Rotated Pole Mapping
=====================

This example uses several visualisation methods to achieve an array of
differing images, including:

 * Visualisation of point based data
 * Contouring of point based data
 * Block plot of contiguous bounded data
 * Non native projection and a Natural Earth shaded relief image underlay

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import iris
import iris.analysis.cartography
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # Load some test data.
    fname = iris.sample_data_path("rotated_pole.nc")
    air_pressure = iris.load_cube(fname)

    # Plot #1: Point plot showing data values & a colorbar
    plt.figure()
    points = qplt.points(air_pressure, c=air_pressure.data)
    cb = plt.colorbar(points, orientation="horizontal")
    cb.set_label(air_pressure.units)
    plt.gca().coastlines()
    iplt.show()

    # Plot #2: Contourf of the point based data
    plt.figure()
    qplt.contourf(air_pressure, 15)
    plt.gca().coastlines()
    iplt.show()

    # Plot #3: Contourf overlayed by coloured point data
    plt.figure()
    qplt.contourf(air_pressure)
    iplt.points(air_pressure, c=air_pressure.data)
    plt.gca().coastlines()
    iplt.show()

    # For the purposes of this example, add some bounds to the latitude
    # and longitude
    air_pressure.coord("grid_latitude").guess_bounds()
    air_pressure.coord("grid_longitude").guess_bounds()

    # Plot #4: Block plot
    plt.figure()
    plt.axes(projection=ccrs.PlateCarree())
    iplt.pcolormesh(air_pressure)
    plt.gca().stock_img()
    plt.gca().coastlines()
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests map creation.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np
import numpy.testing as np_testing
import cartopy.crs as ccrs

import iris
import iris.coord_systems
import iris.cube
import iris.tests.stock

# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import matplotlib.pyplot as plt
    import iris.plot as iplt


# A specific cartopy Globe matching the iris RotatedGeogCS default.
_DEFAULT_GLOBE = ccrs.Globe(
    semimajor_axis=6371229.0, semiminor_axis=6371229.0, ellipse=None
)


@tests.skip_plot
@tests.skip_data
class TestBasic(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.tests.stock.realistic_4d()

    def test_contourf(self):
        cube = self.cube[0, 0]
        iplt.contourf(cube)
        self.check_graphic()

    def test_pcolor(self):
        cube = self.cube[0, 0]
        iplt.pcolor(cube)
        self.check_graphic()

    def test_unmappable(self):
        cube = self.cube[0, 0]
        cube.coord("grid_longitude").standard_name = None
        iplt.contourf(cube)
        self.check_graphic()

    def test_default_projection_and_extent(self):
        self.assertEqual(
            iplt.default_projection(self.cube),
            ccrs.RotatedPole(357.5 - 180, 37.5, globe=_DEFAULT_GLOBE),
        )

        np_testing.assert_array_almost_equal(
            iplt.default_projection_extent(self.cube),
            (3.59579163e02, 3.59669159e02, -1.28250003e-01, -3.82499993e-02),
            decimal=3,
        )


@tests.skip_data
@tests.skip_plot
class TestUnmappable(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        src_cube = iris.tests.stock.global_pp()

        # Make a cube that can't be located on the globe.
        cube = iris.cube.Cube(src_cube.data)
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(96, dtype=np.float32) * 100, long_name="x", units="m"
            ),
            1,
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(73, dtype=np.float32) * 100, long_name="y", units="m"
            ),
            0,
        )
        cube.standard_name = "air_temperature"
        cube.units = "K"
        self.cube = cube

    def test_simple(self):
        iplt.contourf(self.cube, coords=["y", "x"])
        self.check_graphic()


@tests.skip_data
@tests.skip_plot
class TestMappingSubRegion(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        cube_path = tests.get_data_path(
            ("PP", "aPProt1", "rotatedMHtimecube.pp")
        )
        cube = iris.load_cube(cube_path)[0]
        # make the data smaller to speed things up.
        self.cube = cube[::10, ::10]

    def test_simple(self):
        # First sub-plot
        plt.subplot(221)
        plt.title("Default")
        iplt.contourf(self.cube)
        plt.gca().coastlines("110m")

        # Second sub-plot
        plt.subplot(222, projection=ccrs.Mollweide(central_longitude=120))
        plt.title("Molleweide")
        iplt.contourf(self.cube)
        plt.gca().coastlines("110m")

        # Third sub-plot (the projection part is redundant, but a useful
        # test none-the-less)
        ax = plt.subplot(223, projection=iplt.default_projection(self.cube))
        plt.title("Native")
        iplt.contour(self.cube)
        ax.coastlines("110m")

        # Fourth sub-plot
        ax = plt.subplot(2, 2, 4, projection=ccrs.PlateCarree())
        plt.title("PlateCarree")
        iplt.contourf(self.cube)
        ax.coastlines("110m")

        self.check_graphic()

    def test_default_projection_and_extent(self):
        self.assertEqual(
            iplt.default_projection(self.cube),
            ccrs.RotatedPole(357.5 - 180, 37.5, globe=_DEFAULT_GLOBE),
        )

        np_testing.assert_array_almost_equal(
            iplt.default_projection_extent(self.cube),
            (313.01998901, 391.11999512, -22.48999977, 24.80999947),
        )


@tests.skip_data
@tests.skip_plot
class TestLowLevel(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.tests.stock.global_pp()
        self.few = 4
        self.few_levels = list(range(280, 300, 5))
        self.many_levels = np.linspace(
            self.cube.data.min(), self.cube.data.max(), 40
        )

    def test_simple(self):
        iplt.contour(self.cube)
        self.check_graphic()

    def test_params(self):
        iplt.contourf(self.cube, self.few)
        self.check_graphic()

        iplt.contourf(self.cube, self.few_levels)
        self.check_graphic()

        iplt.contourf(self.cube, self.many_levels)
        self.check_graphic()

    def test_keywords(self):
        iplt.contourf(self.cube, levels=self.few_levels)
        self.check_graphic()

        iplt.contourf(self.cube, levels=self.many_levels, alpha=0.5)
        self.check_graphic()


@tests.skip_data
@tests.skip_plot
class TestBoundedCube(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.tests.stock.global_pp()
        # Add some bounds to this data (this will actually make the bounds
        # invalid as they will straddle the north pole and overlap on the
        # dateline, but that doesn't matter for this test.)
        self.cube.coord("latitude").guess_bounds()
        self.cube.coord("longitude").guess_bounds()

    def test_pcolormesh(self):
        # pcolormesh can only be drawn in native coordinates (or more
        # specifically, in coordinates that don't wrap).
        plt.axes(projection=ccrs.PlateCarree(central_longitude=180))
        iplt.pcolormesh(self.cube)
        self.check_graphic()

    def test_grid(self):
        iplt.outline(self.cube)
        self.check_graphic()

    def test_default_projection_and_extent(self):
        self.assertEqual(
            iplt.default_projection(self.cube), ccrs.PlateCarree()
        )
        np_testing.assert_array_almost_equal(
            iplt.default_projection_extent(self.cube),
            [0.0, 360.0, -89.99995422, 89.99998474],
        )
        np_testing.assert_array_almost_equal(
            iplt.default_projection_extent(
                self.cube, mode=iris.coords.BOUND_MODE
            ),
            [-1.875046, 358.124954, -90, 90],
        )


@tests.skip_data
@tests.skip_plot
class TestLimitedAreaCube(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        cube_path = tests.get_data_path(("PP", "aPProt1", "rotated.pp"))
        self.cube = iris.load_cube(cube_path)[::20, ::20]
        self.cube.coord("grid_latitude").guess_bounds()
        self.cube.coord("grid_longitude").guess_bounds()

    def test_pcolormesh(self):
        iplt.pcolormesh(self.cube)
        self.check_graphic()

    def test_grid(self):
        iplt.pcolormesh(self.cube, facecolors="none", edgecolors="blue")
        # the result is a graphic which has coloured edges. This is a mpl bug,
        # see https://github.com/matplotlib/matplotlib/issues/1302
        self.check_graphic()

    def test_outline(self):
        iplt.outline(self.cube)
        self.check_graphic()

    def test_scatter(self):
        iplt.points(self.cube)
        plt.gca().coastlines("110m")
        self.check_graphic()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.analysis.Nearest`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.analysis import Nearest


def create_scheme(mode=None):
    kwargs = {}
    if mode is not None:
        kwargs["extrapolation_mode"] = mode
    return Nearest(**kwargs)


class Test___init__(tests.IrisTest):
    def test_invalid(self):
        with self.assertRaisesRegex(ValueError, "Extrapolation mode"):
            Nearest("bogus")


class Test_extrapolation_mode(tests.IrisTest):
    def check_mode(self, mode):
        scheme = create_scheme(mode)
        self.assertEqual(scheme.extrapolation_mode, mode)

    def test_default(self):
        scheme = Nearest()
        self.assertEqual(scheme.extrapolation_mode, "extrapolate")

    def test_extrapolate(self):
        self.check_mode("extrapolate")

    def test_nan(self):
        self.check_mode("nan")

    def test_error(self):
        self.check_mode("error")

    def test_mask(self):
        self.check_mode("mask")

    def test_nanmask(self):
        self.check_mode("nanmask")


class Test_interpolator(tests.IrisTest):
    def check_mode(self, mode=None):
        scheme = create_scheme(mode)

        # Check that calling `scheme.interpolator(...)` returns an
        # instance of RectilinearInterpolator which has been created
        # using the correct arguments.
        with mock.patch(
            "iris.analysis.RectilinearInterpolator",
            return_value=mock.sentinel.interpolator,
        ) as ri:
            interpolator = scheme.interpolator(
                mock.sentinel.cube, mock.sentinel.coords
            )
        if mode is None:
            expected_mode = "extrapolate"
        else:
            expected_mode = mode
        ri.assert_called_once_with(
            mock.sentinel.cube, mock.sentinel.coords, "nearest", expected_mode
        )
        self.assertIs(interpolator, mock.sentinel.interpolator)

    def test_default(self):
        self.check_mode()

    def test_extrapolate(self):
        self.check_mode("extrapolate")

    def test_nan(self):
        self.check_mode("nan")

    def test_error(self):
        self.check_mode("error")

    def test_mask(self):
        self.check_mode("mask")

    def test_nanmask(self):
        self.check_mode("nanmask")


class Test_regridder(tests.IrisTest):
    def check_mode(self, mode=None):
        scheme = create_scheme(mode)

        # Ensure that calling the regridder results in an instance of
        # RectilinearRegridder being returned, which has been created with
        # the expected arguments.
        with mock.patch(
            "iris.analysis.RectilinearRegridder",
            return_value=mock.sentinel.regridder,
        ) as rr:
            regridder = scheme.regridder(
                mock.sentinel.src_grid, mock.sentinel.tgt_grid
            )

        expected_mode = "extrapolate" if mode is None else mode
        rr.assert_called_once_with(
            mock.sentinel.src_grid,
            mock.sentinel.tgt_grid,
            "nearest",
            expected_mode,
        )
        self.assertIs(regridder, mock.sentinel.regridder)

    def test_default(self):
        self.check_mode()

    def test_extrapolate(self):
        self.check_mode("extrapolate")

    def test_nan(self):
        self.check_mode("nan")

    def test_error(self):
        self.check_mode("error")

    def test_mask(self):
        self.check_mode("mask")

    def test_nanmask(self):
        self.check_mode("nanmask")


if __name__ == "__main__":
    tests.main()
